unit BackPropNet;
{$R-}

interface

uses
  SysUtils, Classes, Controls, Dialogs, Forms, Math, DsgnIntf, TypInfo;

const FileSignature: string = 'neuralnet';

type
  PExtArray = ^TExtArray;
  TExtArray = array[0..0] of Extended;
  TRealArray = class
  private
    ArrayPtr: PExtArray;
    { property data }
    FCapacity: Integer;
    FCount: Integer;
    { property methods }
    function Get(Index: Integer): Extended;
    procedure Put(Index: Integer; Value: Extended);
    procedure SetCapacity(Value: Integer);
  public
    constructor Create(cCount: Integer); virtual;
    destructor Destroy; override;
    function Add(Value: Extended): Integer; virtual;
    procedure Delete(Index: Integer); virtual;
    procedure DeleteAll; virtual;
    procedure Insert(Index: Integer; Value: Extended); virtual;
    property Capacity: Integer read FCapacity write SetCapacity;
    property Count: Integer read FCount;
    property Items[index: Integer]: Extended read Get write Put; default;
  end;

  TPattern = class
  private
    { property data }
    FInputs: TRealArray;
    FOutputs: TRealArray;
  protected
    procedure Assign(Source: TPattern); virtual;
  public
    constructor Create(cInputNo, cOutputNo: Word); virtual;
    destructor Destroy; override;
    property Inputs: TRealArray read FInputs;
    property Outputs: TRealArray read FOutputs;
  end;

const  { file modes }
  fmAdd = $0050;
  fmReplace = $0060;

type
  TSet = class(TList)
  private
    { property data }
    FPatternNo: Integer;           { number of patterns stored in set }
    FInputNo: Integer;             { number of input values stored in patterns }
    FOutputNo: Integer;            { number of output values stored in patterns }
    { property methods }
    function Get(Index: Integer): TPattern;
    procedure SetInputNo(const Value: Integer);
    procedure SetOutputNo(const Value: Integer);
    procedure SetPatternNo(const Value: Integer);
  public
    constructor Create(cPatternNo, cInputNo, cOutputNo: Word); virtual;
    destructor Destroy; override;
    { Adds assigned pattern to set and returns it's position.
      Creates a new TPattern- object and copies all values. }
    function Add(Value: TPattern): Integer; virtual;
    procedure Assign(Source: TSet); virtual;
    { deletes and releases one/ all pattern(s) }
    procedure Delete(Index: Integer); virtual;
    procedure DeleteAll; virtual;
    { Insertes assigned pattern in set at position 'Index'.
      Creates a new TPattern- object and copies all values. }
    procedure Insert(Index: Integer; Value: TPattern); virtual;
    { stream methods }
    procedure AddFromStream(Stream: TStream); virtual;
    procedure AddToStream(Stream: TStream); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    { file methods }
    procedure LoadFromFile(const FileName: string; Mode: Word); virtual;
    procedure SaveToFile(const FileName: string; Mode: Word); virtual;

    property InputNo: Integer read FInputNo write SetInputNo;
    property Items[Index: Integer]: TPattern read Get; default;
    property OutputNo: Integer read FOutputNo write SetOutputNo;
    property PatternNo: Integer read FPatternNo write SetPatternNo;
  end;


  { TNeuron }
  TNeuron = class;
  TFunct = function: Extended of object;
  TFunctTypeProp = (fpUser, fpScalProd, fpSigmaPi);
  TFunctTypeAct  = (faUser, faLinear, faBSB, faDMA, faHopfield);
  TFunctTypeOutp = (foUser, foFermi, foTanh, foSquashCos, foArcTan, foLinear,
                    foThreshold, foThresholdLin, foHardLimit);
  TLearnParam  = (lpWeights, lpBias, lpSlope);
  TLearnParams = set of TLearnParam;
  TUserProc  = procedure(Sender: TNeuron; var Result: Extended) of object;

  TNeuron = class
  private
    { property data }
    FInputNo: Integer;             { number of inputs }
    FInputs: TRealArray;            { input values }

    FAdaptWeights: TRealArray;      { storage of modified adaptation constants }
    FAdaptBias: Extended;
    FAdaptSlope: Extended;

    FLearnParams: TLearnParams;    { training of weights, bias, slope? }
    FParamsMax: Extended;            { range of weights, bias, slope }
    FParamsMin: Extended;

    FActivityMin: Extended;          { range of activity used by DMA- activation function }
    FActivityMax: Extended;
    FBasicValue: Extended;           { basic value of activity,
                                     used by BSB- and DMA- activation functions }
    FDecline: Extended;              { declination of BSB- and DMA- activation functions }
    FScale: Extended;                { scale of activation functions }
    FOutputMax: Extended;            { range of output functions }
    FOutputMin: Extended;

    FWeights: TRealArray;           { weights }
    FWeightsDelta: TRealArray;      { delta of weights }
    FWeightsDeltaOld: TRealArray;   { delta of last step }
    FWeightsDelta2Old: TRealArray;  { delta of step before last }

    FBias: Extended;                 { bias of output functions }
    FBiasDelta: Extended;            { delta of bias }
    FBiasDeltaOld: Extended;         { delta of last step }
    FBiasDelta2Old: Extended;        { delta of step before last }

    FSlope: Extended;                { slope of output functions }
    FSlopeDelta: Extended;           { delta of slope }
    FSlopeDeltaOld: Extended;        { delta of last step }
    FSlopeDelta2Old: Extended;       { delta of step before last }

    FNetInput: Extended;             { effective input }
    FActivity: Extended;             { activity }
    FOutput: Extended;               { output value }
    FOutputDeriv: Extended;          { derivative of output value }

    FTarget: Extended;               { target value }
    FDelta: Extended;                { delta value }
    FError: Extended;                { error = abs(output-target) }

    FActivityAll: Extended;          { total activity for entire pattern set }
    FOutputAll: Extended;            { total output for entire pattern set }
    FPowerAll: Extended;             { total power (activity^2) for entire pattern set }

    FPropFunction: TFunctTypeProp;    { propagation rule }
    FActFunction: TFunctTypeAct;      { activation function }
    FOutpFunction: TFunctTypeOutp;    { output function }

    { events }
    FOnUserPropFunction: TUserProc;      { user propagation rule }
    FOnUserActFunction: TUserProc;       { user activation function }
    FOnUserOutpFunction: TUserProc;      { user output function }
    FOnUserOutpDerivFunction: TUserProc; { user output function derivative}

    { procedural help variables }
    FFunctProp: TFunct;
    FFunctAct: TFunct;
    FFunctOutp: TFunct;
    FFunctOutpDeriv: TFunct;

    { property methods }
    procedure SetActFunction(const Value: TFunctTypeAct);
    procedure SetInputNo(const Value: Integer);
    procedure SetLearnParams(const Value: TLearnParams);
    procedure SetOutpFunction(const Value: TFunctTypeOutp);
    procedure SetParamsRange(Index: Integer; const Value: Extended);
    procedure SetPropFunction(const Value: TFunctTypeProp);
    procedure SetOnUserActFunction(Value: TUserProc);
    procedure SetOnUserOutpFunction(Value: TUserProc);
    procedure SetOnUserPropFunction(Value: TUserProc);
  protected
    { computes random values in interval (ParamsMax,ParamsMin) }
    function GetRandom: Extended; virtual;
    { activation functions }
    function FunctActBSB: Extended;
    function FunctActDMA: Extended;
    function FunctActHopfield: Extended;
    function FunctActLinear: Extended;
    function FunctActUser: Extended;
    { output functions }
    function FunctOutpArcTan: Extended;
    function FunctOutpFermi: Extended;
    function FunctOutpHardLimit: Extended;
    function FunctOutpLinear: Extended;
    function FunctOutpSquashCos: Extended;
    function FunctOutpTanh: Extended;
    function FunctOutpThreshold: Extended;
    function FunctOutpThresholdLin: Extended;
    function FunctOutpUser: Extended;
    { output function derivatives }
    function FunctOutpArcTanDeriv: Extended;
    function FunctOutpFermiDeriv: Extended;
    function FunctOutpLinearDeriv: Extended;
    function FunctOutpSquashCosDeriv: Extended;
    function FunctOutpTanhDeriv: Extended;
    function FunctOutpUserDeriv: Extended;
    { propagation functions }
    function FunctPropScalProd: Extended;
    function FunctPropSigmaPi: Extended;
    function FunctPropUser: Extended;
  public
    constructor Create(cInputNo: Word); virtual;
    destructor Destroy; override;
    { Destination.Assign(Source) -> assignes all relevant properties of
      source neuron to destination neuron }
    procedure Assign(Source: TNeuron); virtual;
    { calculates 'state' (all outputs) of neuron according to current properties }
    procedure CalcState; virtual;
    { decrements/ increments input number of neuron and updates all affected parameters }
    procedure DecInputNo(Index: Integer); virtual;
    procedure IncInputNo(Index: Integer); virtual;
    { reset of several values according to setted properties }
    procedure ResetAdapts(const AdaptConst: Word); virtual;
    procedure ResetOutputs; virtual;
    procedure ResetParams; virtual;

    property ActFunction: TFunctTypeAct read FActFunction write SetActFunction;
    property Activity: Extended read FActivity;
    property ActivityAll: Extended read FActivityAll;
    property ActivityMax: Extended read FActivityMax write FActivityMax;
    property ActivityMin: Extended read FActivityMin write FActivityMin;
    property AdaptBias: Extended read FAdaptBias write FAdaptBias;
    property AdaptSlope: Extended read FAdaptSlope write FAdaptSlope;
    property AdaptWeights: TRealArray read FAdaptWeights write FAdaptWeights;
    property BasicValue: Extended read FBasicValue write FBasicValue;
    property Bias: Extended read FBias write FBias;
    property BiasDelta: Extended read FBiasDelta write FBiasDelta;
    property BiasDeltaOld: Extended read FBiasDeltaOld write FBiasDeltaOld;
    property BiasDelta2Old: Extended read FBiasDelta2Old write FBiasDelta2Old;
    property Decline: Extended read FDecline write FDecline;
    property Delta: Extended read FDelta write FDelta;
    property Error: Extended read FError;
    property InputNo: Integer read FInputNo write SetInputNo;
    property Inputs: TRealArray read FInputs write FInputs;
    property LearnParams: TLearnParams read FLearnParams write SetLearnParams;
    property NetInput: Extended read FNetInput;
    property OnUserActFunction: TUserProc read FOnUserActFunction write SetOnUserActFunction;
    property OnUserOutpFunction: TUserProc read FOnUserOutpFunction write SetOnUserOutpFunction;
    property OnUserOutpDerivFunction: TUserProc read FOnUserOutpDerivFunction write FOnUserOutpDerivFunction;
    property OnUserPropFunction: TUserProc read FOnUserPropFunction write SetOnUserPropFunction;
    property OutpFunction: TFunctTypeOutp read FOutpFunction write SetOutpFunction;
    property Output: Extended read FOutput;
    property OutputAll: Extended read FOutputAll;
    property OutputDeriv: Extended read FOutputDeriv;
    property OutputMax: Extended read FOutputMax write FOutputMax;
    property OutputMin: Extended read FOutputMin write FOutputMin;
    property ParamsMax: Extended index 1 read FParamsMax write SetParamsRange;
    property ParamsMin: Extended index 2 read FParamsMin write SetParamsRange;
    property PowerAll: Extended read FPowerAll;
    property PropFunction: TFunctTypeProp read FPropFunction write SetPropFunction;
    property Scale: Extended read FScale write FScale;
    property Slope: Extended read FSlope write FSlope;
    property SlopeDelta: Extended read FSlopeDelta write FSlopeDelta;
    property SlopeDeltaOld: Extended read FSlopeDeltaOld write FSlopeDeltaOld;
    property SlopeDelta2Old: Extended read FSlopeDelta2Old write FSlopeDelta2Old;
    property Target: Extended read FTarget write FTarget;
    property Weights: TRealArray read FWeights write FWeights;
    property WeightsDelta: TRealArray read FWeightsDelta write FWeightsDelta;
    property WeightsDeltaOld: TRealArray read FWeightsDeltaOld write FWeightsDeltaOld;
    property WeightsDelta2Old: TRealArray read FWeightsDelta2Old write FWeightsDelta2Old;
  end;


  { TLayer }
  TBackPropNet = class;
  TLayer = class(TPersistent)
  private
    List: TList;                   { stores neurons of layer }
    { property data }
    FNet: TBackPropNet;            { points to neural net containing this layer }
    FSelected: Boolean;            { only used by property editor }
    FNeuronNo: Integer;               { number of neurons stored in layer }
    FPrevLayer: TLayer;            { previous layer in net }
    FNextLayer: TLayer;            { next layer in net }
    { property methods }
    function Get(Index: Integer): TNeuron;
    function GetActFunction: TFunctTypeAct;
    function GetBias: Extended;
    function GetOnUserActFunction: TUserProc;
    function GetOnUserOutpFunction: TUserProc;
    function GetOnUserOutpDerivFunction: TUserProc;
    function GetOnUserPropFunction: TUserProc;
    function GetOutpFunction: TFunctTypeOutp;
    function GetOutputMax: Extended;
    function GetOutputMin: Extended;
    function GetPropFunction: TFunctTypeProp;
    function GetScale: Extended;
    function GetSlope: Extended;
    procedure SetActFunction(const Value: TFunctTypeAct);
    procedure SetBias(const Value: Extended);
    procedure SetNeuronNo(const Value: Integer);
    procedure SetOnUserActFunction(Value: TUserProc);
    procedure SetOnUserOutpFunction(Value: TUserProc);
    procedure SetOnUserOutpDerivFunction(Value: TUserProc);
    procedure SetOnUserPropFunction(Value: TUserProc);
    procedure SetOutpFunction(const Value: TFunctTypeOutp);
    procedure SetOutputMax(const Value: Extended);
    procedure SetOutputMin(const Value: Extended);
    procedure SetPropFunction(const Value: TFunctTypeProp);
    procedure SetScale(const Value: Extended);
    procedure SetSlope(const Value: Extended);
  public
    constructor Create(cNeuronNo, cInputNo: Word); virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    { deletes neuron at position 'Index' }
    procedure DeleteNeuron(Index: Integer); virtual;
    { executes and calculates layer }
    procedure FeedForward; virtual;
    { insertes neuron at position 'Index' }
    procedure InsertNeuron(Index: Integer); virtual;
    { loads/ saves layer from/ to assigned Stream }
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;

    property Items[Index: Integer]: TNeuron read Get; default;
    property Net: TBackPropNet read FNet write FNet;
    property NextLayer: TLayer read FNextLayer write FNextLayer;
    property PrevLayer: TLayer read FPrevLayer write FPrevLayer;
    property Selected: Boolean read FSelected write FSelected;
  published
    property ActFunction: TFunctTypeAct read GetActFunction write SetActFunction default faLinear;
    property Bias: Extended read GetBias write SetBias;
    property NeuronNo: Integer read FNeuronNo write SetNeuronNo nodefault;
    property OnUserActFunction: TUserProc read GetOnUserActFunction write SetOnUserActFunction;
    property OnUserOutpFunction: TUserProc read GetOnUserOutpFunction write SetOnUserOutpFunction;
    property OnUserOutpDerivFunction: TUserProc read GetOnUserOutpDerivFunction write SetOnUserOutpDerivFunction;
    property OnUserPropFunction: TUserProc read GetOnUserPropFunction write SetOnUserPropFunction;
    property OutpFunction: TFunctTypeOutp read GetOutpFunction write SetOutpFunction default foFermi;
    property OutputMax: Extended read GetOutputMax write SetOutputMax;
    property OutputMin: Extended read GetOutputMin write SetOutputMin;
    property PropFunction: TFunctTypeProp read GetPropFunction write SetPropFunction default fpScalProd;
    property Scale: Extended read GetScale write SetScale;
    property Slope: Extended read GetSlope write SetSlope;
  end;


  { TBackPropNet }
  TExitCode = (exBreak, exCycle, exError);
  TExitEvent = procedure(Sender: TBackPropNet; const ExitCode: TExitCode) of object;
  TIntervalEvent = procedure(Sender: TBackPropNet) of object;
  TLearnMethod = (lmEpochal, lmDirect);

  TBackPropNet = class(TComponent)
  private
    List: TList;                      { stores layers of net }
    { property data }
    FBreak: Boolean;                  { interrupts learn process/ recall }

    FInputNo: Integer;                { number of inputs of net }
    FLayerNo: Integer;                { number of layers }
    FPatternSet: TSet;                { subset of patterns to be learned in one epoch }
    FRecallSet: TSet;                 { subset of patterns to be recalled in one operation }

    FAdaptConst: Integer;             { adaptation constant for recursive mean estimation }
    FAdaptConstStep: Extended;          { variable adaptation constant +- step }
    FAdaptConstFactor: Extended;        { variable adaptation constant */ factor }

    FLearnRate: Extended;               { learn rate of net}
    FLearnRateAdapt: Extended;          { storage of modified learn rate}
    FLearnRateDec: Extended;            { learn rate decrease factor }
    FLearnRateInc: Extended;            { learn rate increase factor }
    FMomentum: Extended;                { momentum term }

    FLearnMethod: TLearnMethod;       { epochal or direct learn step? }
    FLearnParams: TLearnParams;       { learning weights, bias, slope? }
    FParamsMax: Extended;               { range of weights, bias, slope }
    FParamsMin: Extended;

    FErrorLimit: Extended;              { error goal -> exit learn process if GlobalError' is this low }
    FCycleLimit: Integer;             { maximum cycles allowed in learn mode }
    FInterval: Integer;               { output interval }

    FGlobalError: Extended;             { sum squared error of NeuroNet for entire pattern set }
    FGlobalErrorOld: Extended;          { sum squared error of last step }
    FGlobalError2Old: Extended;         { sum squared error of step before last }
    FCycles: Integer;                 { number of cycles passed }

    { events }
    FOnInterval: TIntervalEvent;      { Event is released:
                                         # every time 'Interval'* Cycles are passed in learn mode,
                                         # after each input vector is passed in recall mode. }
    FOnExit: TExitEvent;              { Event is released:
                                         # if either 'CycleLimit' or 'ErrorLimit' is reached by 'GlobalError',
                                           or 'CycleLimit' respectively in learn mode,
                                         # if 'Break' is setted to 'true' in learn mode or recall mode,
                                         # if entire recall set is passed in recall mode.
                                        The cause of exit is indicated by the parameter 'ExitCode' and can
                                        hold either exBreak, exCycle or exError. }
    { property methods }
    function Get(Index: Integer): TLayer;
    function GetLayer: TLayer;
    procedure SetAdaptConst(const Value: Integer);
    procedure SetAdaptConstFactor(const Value: Extended);
    procedure SetAdaptConstStep(const Value: Extended);
    procedure SetInputNo(const Value: Integer);
    procedure SetInterval(const Value: Integer);
    procedure SetLayer(const Value: TLayer);
    procedure SetLayerNo(const Value: Integer);
    procedure SetLearnParams(const Value: TLearnParams);
    procedure SetLearnRate(const Value: Extended);
    procedure SetLearnRateDec(const Value: Extended);
    procedure SetLearnRateInc(const Value: Extended);
    procedure SetMomentum(const Value: Extended);
    procedure SetParamsMax(const Value: Extended);
    procedure SetParamsMin(const Value: Extended);
    procedure SetPatternSet(Value: TSet);
    procedure SetRecallSet(Value: TSet);
  protected
    { Backward propagation of output error.
      Update of weights, bias and slope after each pattern is passed (in direct learn mode). }
    procedure BackProp; virtual;
    { executes and calculates the net }
    procedure FeedForward; virtual;
    { estimates sum squared error of output layer for each pattern }
    function GetPatternError: Extended; virtual;
    { update of weights, bias and slope after entire pattern set is passed (in epochal learn mode) }
    procedure SetEpochWeights; virtual;
    { assignes input values (PatternSet.Inputs/ RecallSet.Inputs) to input layer }
    procedure SetInputs(InpVector: TRealArray); virtual;
    { assignes outputs of net to RecallSet.Outputs in recall mode }
    procedure SetOutputs(OutpVector: TRealArray); virtual;
    { assignes target values (PatternSet.Outputs) to output layer }
    procedure SetTargets(TargVector: TRealArray); virtual;
    { stream methods used by procedures 'DefineProperties'
      'SaveToFile' and 'LoadFromFile' }
    procedure LoadLayersFromStream(Stream: TStream); virtual;
    procedure SaveLayersToStream(Stream: TStream); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    { defines "false" properties 'Layers', 'Pattern' and 'Recall'
      to enable storage of all layers, PatternSet and RecallSet in stream }
    procedure DefineProperties(Filer: TFiler); override;
    { deletes layer at position 'Index' }
    procedure DeleteLayer(Index: Integer); virtual;
    { deletes neuron in 'Net[Index][NeuronIndex]' }
    procedure DeleteNeuron(Index, NeuronIndex: Integer); virtual;
    { insertes layer at position 'Index' containing 'Size' neurons }
    procedure InsertLayer(Index, Size: Integer); virtual;
    { insertes neuron at 'Net[Index][NeuronIndex]' }
    procedure InsertNeuron(Index, NeuronIndex: Integer); virtual;
    { one learn step -> learnes net for entire PatternSet }
    procedure Learn; virtual;
    { propagation of net using data in RecallSet }
    procedure Recall; virtual;
    { resets several values according to setted properties }
    procedure ResetAdapts; virtual;
    procedure ResetAll; virtual;
    procedure ResetOutputs; virtual;
    procedure ResetParams; virtual;
    { stream methods used by procedures 'DefineProperties'
      'SaveToFile' and 'LoadFromFile' }
    procedure LoadConfigFromStream(Stream: TStream); virtual;
    procedure LoadNetFromStream(Stream: TStream); virtual;
    procedure LoadParamsFromStream(Stream: TStream); virtual;
    procedure SaveConfigToStream(Stream: TStream); virtual;
    procedure SaveNetToStream(Stream: TStream); virtual;
    procedure SaveParamsToStream(Stream: TStream); virtual;
    { Loads/ saves data from/ to a file spezified by 'FileName'. According to
      file extension different data will be processed:
         *.cfg: configuration of net (LayerNo,..);
         *.lrn: learn parameters of net (Weights, Bias, Slope);
         *.ptn: PatternSet;   *.rcl: RecallSet; }
    procedure LoadFromFile(const FileName: string); virtual;
    procedure SaveToFile(const FileName: string); virtual;

    property Break: Boolean read FBreak write FBreak;
    property Cycles: Integer read FCycles;
    property GlobalError: Extended read FGlobalError;
    property Items[Index: Integer]: TLayer read Get; default;
    property LearnRateAdapt: Extended read FLearnRateAdapt;
  published
    property AdaptConst: Integer read FAdaptConst write SetAdaptConst default 0;
    property AdaptConstFactor: Extended read FAdaptConstFactor write SetAdaptConstFactor;
    property AdaptConstStep: Extended read FAdaptConstStep write SetAdaptConstStep;
    property CycleLimit: Integer read FCycleLimit write FCycleLimit default 640;
    property ErrorLimit: Extended read FErrorLimit write FErrorLimit;
    property InputNo: Integer read FInputNo write SetInputNo nodefault;
    property Interval: Integer read FInterval write SetInterval default 1;
    { 'Layer' just reserves a field in object inspector to access a property editor
      'TLayerProperty'. It's used only by property editors during development time. }
    property Layer: TLayer read GetLayer write SetLayer stored False;
    property LayerNo: Integer read FLayerNo write SetLayerNo nodefault;
    property LearnMethod: TLearnMethod read FLearnMethod write FLearnMethod default lmEpochal;
    property LearnParams: TLearnParams read FLearnParams write SetLearnParams default [lpWeights, lpBias];
    property LearnRate: Extended read FLearnRate write SetLearnRate;
    property LearnRateDec: Extended read FLearnRateDec write SetLearnRateDec;
    property LearnRateInc: Extended read FLearnRateInc write SetLearnRateInc;
    property Momentum: Extended read FMomentum write SetMomentum;
    property OnExit: TExitEvent read FOnExit write FOnExit;
    property OnInterval: TIntervalEvent read FOnInterval write FOnInterval;
    property ParamsMax: Extended read FParamsMax write SetParamsMax;
    property ParamsMin: Extended read FParamsMin write SetParamsMin;
    property PatternSet: TSet read FPatternSet write SetPatternSet stored false;
    property RecallSet: TSet read FRecallSet write SetRecallSet stored false;
  end;


  { TPatternDialog }
  TSetType = (stPatternSet, stRecallSet);
  TPatternDialog = class(TComponent)
  private
    FSetType: TSetType;           { Edit Pattern- or RecallSet? }
    FNeuralNet: TBackPropNet;     { points to net }
    FPatternSet: TSet;            { points to Set }
    FTitle: string;
    procedure SetSetType(Value: TSetType);
    procedure SetPatternSet(const Value: TSet);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: Boolean; virtual;
    property PatternSet: TSet read FPatternSet write SetPatternSet;
  published
    property NeuralNet: TBackPropNet read FNeuralNet write FNeuralNet;
    property SetType: TSetType read FSetType write SetSetType default stPatternSet;
    property Title: string read FTitle write FTitle;
  end;


  { property editors }
  TPatternProperty = class(TPropertyEditor)
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
  end;

  TLayerIntProperty = class(TOrdinalProperty)
    function AllEqual: Boolean; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  TLayerEnumProperty = class(TEnumProperty)
    function AllEqual: Boolean; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  TLayerFloatProperty = class(TPropertyEditor)
    function AllEqual: Boolean; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  TLayerProperty = class(TClassProperty)
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
  end;

  { component editor }
  TLayerComponentEditor = class(TComponentEditor)
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure Register;

implementation
uses FileDlg, InsertDlg, SelectDlg, PatternDlg;

{ methodes of TIntegerArray }
constructor TRealArray.Create(cCount: Integer);
begin
  inherited Create;
  if cCount = 0 then Exit;
  ArrayPtr := AllocMem(cCount * SizeOf(Extended));
  if (ArrayPtr = nil) then OutOfMemoryError;
  FCount := cCount;
  FCapacity := cCount;
end;

destructor TRealArray.Destroy;
begin
  if (ArrayPtr <> nil) then
    FreeMem(ArrayPtr, 0);
  inherited Destroy;
end;

function TRealArray.Add(Value: Extended): Integer;
begin
  Result := FCount;
  if FCount = FCapacity then begin
    Inc(FCapacity, 8);
    ReallocMem(ArrayPtr, FCapacity * SizeOf(Extended));
  end;
  ArrayPtr^[FCount] := Value;
  Inc(FCount);
end;

procedure TRealArray.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then
    raise ERangeError.Create('Bereichsberschreitung');
  Dec(FCount);
  System.Move(ArrayPtr^[Index + 1], ArrayPtr^[Index],
              (FCount - Index) * SizeOf(Extended));
end;

procedure TRealArray.DeleteAll;
begin
  ReallocMem(ArrayPtr, 0);
  FCapacity := 0;
  FCount := 0;
end;

function TRealArray.Get(Index: Integer): Extended;
begin
  if (Index < 0) or (Index >= FCount) then
    raise ERangeError.Create('Bereichsberschreitung');
  Result := ArrayPtr^[Index];
end;

procedure TRealArray.Insert(Index: Integer; Value: Extended);
begin
  if (Index < 0) or (Index > FCount) then
    raise ERangeError.Create('Bereichsberschreitung');
  if FCount = FCapacity then begin
    Inc(FCapacity, 8);
    ReallocMem(ArrayPtr, FCapacity * SizeOf(Extended));
  end;
  if Index < FCount then
    System.Move(ArrayPtr^[Index], ArrayPtr^[Index + 1],
                (FCount - Index) * SizeOf(Extended));
  ArrayPtr^[Index] := Value;
  Inc(FCount);
end;

procedure TRealArray.Put(Index: Integer; Value: Extended);
begin
  if (Index < 0) or (Index >= FCount) then
    raise ERangeError.Create('Bereichsberschreitung');
  ArrayPtr^[Index] := Value;
end;

procedure TRealArray.SetCapacity(Value: Integer);
begin
  if Value < FCount then
    raise ERangeError.Create('Bereichsberschreitung');
  if Value <> FCapacity then
  begin
    ReallocMem(ArrayPtr, Value * SizeOf(Extended));
    FCapacity := Value;
  end;
end;

{ methods of TPattern }
constructor TPattern.Create(cInputNo, cOutputNo: Word);
begin
  inherited Create;
  FInputs := TRealArray.Create(cInputNo);
  FOutputs := TRealArray.Create(cOutputNo);
end;

destructor TPattern.Destroy;
begin
  FInputs.Free;
  FOutputs.Free;
  inherited Destroy;
end;

procedure TPattern.Assign(Source: TPattern);
var I: Integer;
begin
  with Source do begin
    for I := 0 to (Inputs.Count - 1) do
      Self.FInputs[I] := Inputs[I];
    for I := 0 to (Outputs.Count - 1) do
      Self.FOutputs[I] := Outputs[I];
    end;
end;


{ methods of TSet }
constructor TSet.Create(cPatternNo, cInputNo, cOutputNo: Word);
var I: Integer; Pattern: TPattern;
begin
  inherited Create;
  FInputNo := cInputNo;
  FOutputNo := cOutputNo;
  if cPatternNo = 0 then Exit;
  FPatternNo := cPatternNo;
  Capacity := FPatternNo;
  for I := 0 to (FPatternNo - 1) do begin
    Pattern := TPattern.Create(FInputNo, FOutputNo);
    TList(Self).Add(Pattern);
  end;
end;

destructor TSet.Destroy;
begin
  DeleteAll;
  inherited Destroy;
end;

function TSet.Add(Value: TPattern): Integer;
var Pattern: TPattern;
begin
  if Value.Inputs.Count > FInputNo then
    InputNo := Value.Inputs.Count;
  if Value.Outputs.Count > FOutputNo then
    OutputNo := Value.Outputs.Count;
  Inc(FPatternNo);
  Capacity := FPatternNo;
  Pattern := TPattern.Create(FInputNo, FOutputNo);
  Pattern.Assign(Value);
  Result := TList(Self).Add(Pattern);
end;

procedure TSet.AddFromStream(Stream: TStream);
var
  I, J: Integer;
  InpNoStream, OutpNoStream, PatNoSet: Integer;
  DoubleVal: Extended;
begin
  Stream.Read(I, SizeOf(I));  { read PatternNo }
  if I = 0 then Exit;
  PatNoSet := FPatternNo;
  PatternNo := FPatternNo + I;
  Stream.Read(InpNoStream, SizeOf(InpNoStream));  { read InputNo }
  if InpNoStream > FInputNo then InputNo := InpNoStream;
  Stream.Read(OutpNoStream, SizeOf(OutpNoStream));  { read OutputNo }
  if OutpNoStream > FOutputNo then OutputNo := OutpNoStream;
  for I := PatNoSet to (FPatternNo - 1) do
    with Items[I] do begin
      for J := 0 to (InpNoStream - 1) do begin
        Stream.Read(DoubleVal, SizeOf(DoubleVal));
        Inputs[J] := DoubleVal;
      end;
      for J := 0 to (OutpNoStream - 1) do begin
        Stream.Read(DoubleVal, SizeOf(DoubleVal));
        Outputs[J] := DoubleVal;
      end;
    end;
end;

procedure TSet.AddToStream(Stream: TStream);
var ASet: TSet;
begin
  ASet := TSet.Create(0, 0, 0);
  try
    ASet.Assign(Self);
    ASet.AddFromStream(Stream);
    Stream.Seek(9, soFromBeginning);
    ASet.SaveToStream(Stream);
  finally
    ASet.Free;
  end;
end;

procedure TSet.Assign(Source: TSet);
var I: Integer;
begin
  with Source do begin
    Self.PatternNo := PatternNo;
    Self.InputNo := InputNo;
    Self.OutputNo := OutputNo;
    for I := 0 to (PatternNo - 1) do
      Self[I].Assign(Items[I]);
    end;
end;

procedure TSet.Delete(Index: Integer);
begin
  if (Index >= 0) and (Index < FPatternNo) then begin
    Items[Index].Free;
    TList(Self).Delete(Index);
    Dec(FPatternNo);
  end;
end;

procedure TSet.DeleteAll;
begin
  while Count > 0 do
    Delete(Count - 1);
end;

function TSet.Get(Index: Integer): TPattern;
begin
  Result := TPattern(TList(Self).Items[Index]);
end;

procedure TSet.Insert(Index: Integer; Value: TPattern);
var Pattern: TPattern;
begin
  if (Index >= 0) and (Index <= FPatternNo) then begin
    if Value.Inputs.Count > FInputNo then
      InputNo := Value.Inputs.Count;
    if Value.Outputs.Count > FOutputNo then
      OutputNo := Value.Outputs.Count;
    Inc(FPatternNo);
    Capacity := FPatternNo;
    Pattern := TPattern.Create(FInputNo, FOutputNo);
    Pattern.Assign(Value);
    TList(Self).Insert(Index, Pattern);
  end;
end;

procedure TSet.LoadFromFile(const FileName: string; Mode: Word);
var
  Stream: TStream;
  TestSignature: string[9];
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    Stream.Read(TestSignature[1], 9);
    TestSignature[0] := #9;
    if AnsiCompareText(FileSignature, TestSignature) <> 0 then
      raise Exception.Create('Datei ''' + FileName + ''' ist keine Netzdatei');
    if Mode = fmAdd then AddFromStream(Stream)
                    else LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TSet.LoadFromStream(Stream: TStream);
var
  I, J: Integer;
  DoubleVal: Extended;
begin
  Stream.Read(I, SizeOf(I));
  PatternNo := I;
  Stream.Read(I, SizeOf(I));
  InputNo := I;
  Stream.Read(I, SizeOf(I));
  OutputNo := I;
  for I := 0 to (FPatternNo - 1) do
    with Items[I] do begin
      for J := 0 to (FInputNo - 1) do begin
        Stream.Read(DoubleVal, SizeOf(DoubleVal));
        Inputs[J] := DoubleVal;
      end;
      for J := 0 to (FOutputNo - 1) do begin
        Stream.Read(DoubleVal, SizeOf(DoubleVal));
        Outputs[J] := DoubleVal;
      end;
    end;
end;

procedure TSet.SaveToFile(const FileName: string; Mode: Word);
var
  Stream: TStream;
  TestSignature: string[9];
begin
  if Mode = fmAdd then begin
    if not FileExists(FileName) then
      SaveToFile(FileName, fmReplace)  { create new file if not exists }
    else begin  { add to existing file }
      Stream := TFileStream.Create(FileName, fmOpenReadWrite);
      try
        Stream.Read(TestSignature[1], 9);
        TestSignature[0] := #9;
        if AnsiCompareText(FileSignature, TestSignature) <> 0 then
          raise Exception.Create('Datei ''' + FileName + ''' ist keine Netzdatei');
        AddToStream(Stream);
        Stream.Free;
      except
        on E: Exception do begin
          Stream.Free;
          if MessageDlg(E.Message + #13#10 + 'Die Datei ''' + FileName +
                        ''' enthlt kein gltiges Set.' + #13#10 +
                        'Mchten Sie sie berschreiben?', mtError, [mbYes, mbNo], 0) = mrYes then
            SaveToFile(FileName, fmReplace);  { replace existing invalid file }
        end;
      end;
    end;
  end
  else begin  { replace existing file or create new one }
    Stream := TFileStream.Create(FileName, fmCreate);
    try
      Stream.Write(FileSignature[1], 9);
      SaveToStream(Stream);
    finally
      Stream.Free;
    end;
  end;
end;

procedure TSet.SaveToStream(Stream: TStream);
var I, J: Integer; DoubleVal: Extended;
begin
  Stream.Write(FPatternNo, SizeOf(FPatternNo));
  Stream.Write(FInputNo, SizeOf(FInputNo));
  Stream.Write(FOutputNo, SizeOf(FOutputNo));
  for I := 0 to (FPatternNo - 1) do
    with Items[I] do begin
      for J := 0 to (FInputNo - 1) do begin
        DoubleVal := Inputs[J];
        Stream.Write(DoubleVal, SizeOf(DoubleVal));
      end;
      for J := 0 to (FOutputNo - 1) do begin
        DoubleVal := Outputs[J];
        Stream.Write(DoubleVal, SizeOf(DoubleVal));
      end;
    end;
end;

procedure TSet.SetInputNo(const Value: Integer);
var I: Integer;
begin
  if Value >= 0 then begin
    while FInputNo > Value do begin
      for I := 0 to (FPatternNo - 1) do
        with Items[I].Inputs do Delete(Count - 1);
      Dec(FInputNo);
    end;
    while FInputNo < Value do begin
      for I := 0 to (FPatternNo - 1) do
        Items[I].Inputs.Add(0.0);
      Inc(FInputNo);
    end;
  end;
end;

procedure TSet.SetPatternNo(const Value: Integer);
var Pattern: TPattern;
begin
  if Value >= 0 then begin
    while Count > Value do
      Delete(Count - 1);
    while Count < Value do begin
      Pattern := TPattern.Create(FInputNo, FOutputNo);
      TList(Self).Add(Pattern);
      Inc(FPatternNo);
    end;
  end;
end;

procedure TSet.SetOutputNo(const Value: Integer);
var I: Integer;
begin
  if Value >= 0 then begin
    while FOutputNo > Value do begin
      for I := 0 to (FPatternNo - 1) do
        with Items[I].Outputs do Delete(Count - 1);
      Dec(FOutputNo);
    end;
    while FOutputNo < Value do begin
      for I := 0 to (FPatternNo - 1) do
        Items[I].Outputs.Add(0.0);
      Inc(FOutputNo);
    end;
  end;
end;


{ methods of TNeuron }
constructor TNeuron.Create(cInputNo: Word);
var I: Integer;
begin
  if cInputNo = 0 then
    raise Exception.Create('Kann Neuron nicht erstellen.' + #13#10 +
                           'Das Neuron mu mindestens einen Eingang besitzen');
  inherited Create;
  FInputNo := cInputNo;
  FActivityMax := 1.0;
  FBasicValue := 0.5;
  FDecline := 0.5;
  FLearnParams := [lpWeights, lpBias];
  FOutputMax := 1.0;
  FParamsMax := 0.9;
  FParamsMin := -0.9;
  FScale := 1.0;
  FSlope := 1.0;
  ActFunction := faLinear;
  OutpFunction := foFermi;
  PropFunction := fpScalProd;
  FAdaptWeights := TRealArray.Create(FInputNo);
  FInputs := TRealArray.Create(FInputNo);
  FWeights := TRealArray.Create(FInputNo);
  FWeightsDelta := TRealArray.Create(FInputNo);
  FWeightsDeltaOld := TRealArray.Create(FInputNo);
  FWeightsDelta2Old := TRealArray.Create(FInputNo);
  FBias := GetRandom;
  for I := 0 to (FInputNo - 1) do
    FWeights[I] := GetRandom;
end;

destructor TNeuron.Destroy;
begin
  FInputs.Free;
  FAdaptWeights.Free;
  FWeights.Free;
  FWeightsDelta.Free;
  FWeightsDeltaOld.Free;
  FWeightsDelta2Old.Free;
  inherited Destroy;
end;

procedure TNeuron.Assign(Source: TNeuron);
begin
  with Source do begin
    Self.FParamsMax := ParamsMax;
    Self.FParamsMin := ParamsMin;
    Self.FOutputMax := OutputMax;
    Self.FOutputMin := OutputMin;
    Self.FScale := Scale;
    if AdaptSlope > 1 then
      Self.ResetAdapts(Trunc(AdaptSlope));
    Self.FOnUserPropFunction := OnUserPropFunction;
    Self.FOnUserActFunction := OnUserActFunction;
    Self.FOnUserOutpFunction := OnUserOutpFunction;
    Self.FOnUserOutpDerivFunction := OnUserOutpDerivFunction;
    Self.PropFunction := PropFunction;
    Self.ActFunction := ActFunction;
    Self.OutpFunction := OutpFunction;
    Self.LearnParams := LearnParams;
  end;
end;

procedure TNeuron.CalcState;
begin
  FNetInput := FFunctProp;
  FActivity := FFunctAct;
  FOutput := FFunctOutp;
  if Assigned(FFunctOutpDeriv) then FOutputDeriv := FFunctOutpDeriv;
  FError := Abs(FOutput - FTarget);
  FActivityAll := FActivityAll + FActivity;
  FPowerAll := FPowerAll + Sqr(FActivity);            
  FOutputAll := FOutputAll + FOutput;
end;

procedure TNeuron.DecInputNo(Index: Integer);
begin
  if (Index >= 0) and (Index < FInputNo) then begin
    if FInputNo = 1 then
      raise Exception.Create('Kann Anzahl der Eingnge nicht verringern.' + #13#10 +
                             'Das Neuron mu mindestens einen Eingang besitzen');
    FAdaptWeights.Delete(Index);
    FInputs.Delete(Index);
    FWeights.Delete(Index);
    FWeightsDelta.Delete(Index);
    FWeightsDeltaOld.Delete(Index);
    FWeightsDelta2Old.Delete(Index);
    Dec(FInputNo);
  end;
end;

{ activation functions }
function TNeuron.FunctActBSB: Extended;
begin
  Result := FActivity + FScale * FNetInput;
  Result := Result - FDecline * (FActivity - FBasicValue);
end;

function TNeuron.FunctActDMA: Extended;
var Val: Extended;
begin
  Result := FActivity - FDecline * (FActivity - FBasicValue);
  if FNetInput < 0 then
    Val := FScale * FNetInput * (FActivity - FActivityMin)
  else
    Val := FScale * FNetInput * (FActivityMax - FActivity);
  Result := Result + Val;
end;

function TNeuron.FunctActHopfield: Extended;
begin
  Result := 1.0;
  if FNetInput < 0 then Result := FActivityMin
    else if FNetInput = 0 then Result := FActivity;
end;

function TNeuron.FunctActLinear: Extended;
begin
  Result := FNetInput * FScale;
end;

function TNeuron.FunctActUser: Extended;
begin
  Result := FActivity;
  if Assigned(FOnUserActFunction) then FOnUserActFunction(Self, Result);
end;

{ output functions }
function TNeuron.FunctOutpArcTan: Extended;
var Val: Extended;
begin
  Val := (FOutputMax - FOutputMin) / 2.0;
  Result := ArcTan(FSlope * (FActivity - FBias) / Val);
  Result := Val + FOutputMin + Val * 2.0 / Pi * Result;
end;

function TNeuron.FunctOutpFermi: Extended;
var Val: Extended;
begin
  Val := FOutputMax - FOutputMin;
  Result := -FSlope * (FActivity - FBias) / Val;
  { if exp(Result) > MaxExtended then Result := MaxExtended --> avoid overflow }
  if Result > 11354 then Result := 1E4931
                    else Result := 1 + Exp(Result);
  { avoid division by zero }
  if Result = 0 then Result := Result + MinExtended;
  Result := FOutputMin + Val / Result
end;

function TNeuron.FunctOutpHardLimit: Extended;
begin
  if FSlope = 0 then FSlope := FSlope + 1E-4930;
  Result := (FOutputMax - FOutputMin) / 2.0 / FSlope;
  if (FActivity - FBias) < (-Result) then Result := FOutputMin
    else if (FActivity - FBias) > Result then Result := FOutputMax
      else Result := FSlope * (FActivity - FBias + Result) ;
end;

function TNeuron.FunctOutpLinear: Extended;
begin
  Result := FSlope * (FActivity - FBias);
end;

function TNeuron.FunctOutpSquashCos: Extended;
var Val: Extended;
begin
  Val := (FOutputMax - FOutputMin) / 2.0;
  Result := FSlope * (FActivity - FBias) / Val;
  if Result < (-Pi / 2.0) then Result := FOutputMin
    else if Result > (Pi / 2.0) then Result := FOutputMax
      else begin
             Result := Sin(Result);
             Result := Val + FOutputMin + Val * Result;
           end;
end;

function TNeuron.FunctOutpTanh: Extended;
var Val: Extended;
begin
  Val := (FOutputMax - FOutputMin) / 2.0;
  Result := Tanh(FSlope * (FActivity - FBias) / Val);
  Result := Val + FOutputMin + Val * Result;
end;

function TNeuron.FunctOutpThreshold: Extended;
begin
  if FActivity < FBias then Result := FOutputMin
    else Result := FOutputMax;
end;

function TNeuron.FunctOutpThresholdLin: Extended;
begin
  if FActivity < FBias then Result := FOutputMin
    else Result := FSlope * (FActivity - FBias) + FOutputMin;
end;

function TNeuron.FunctOutpUser: Extended;
begin
  Result := FOutput;
  if Assigned(FOnUserOutpFunction) then FOnUserOutpFunction(Self, Result);
end;

{ output function derivatives }
function TNeuron.FunctOutpArcTanDeriv: Extended;
begin
  Result := (FOutputMax - FOutputMin) / 2.0;
  Result := FSlope * (FActivity - FBias) / Result;
  Result := 1.0 + Result * Result;
  Result := 0.2 + 2.0 / Pi * FSlope / Result;
end;

function TNeuron.FunctOutpFermiDeriv: Extended;
begin
  Result := FOutputMax - FOutputMin;
  Result := -FSlope * (FActivity - FBias) / Result;
  { if exp(Result) > MaxExtended then Result := MaxExtended --> avoid Overflow }
  if Result > 11354 then Result := 1E4931
                    else Result := Exp(Result);
  { avoid division by zero }
  if Result = 0 then Result := Result + MinExtended;
  Result := 1.0 / Result + Result + 2.0;
  Result := 0.1 + FSlope / Result;
end;

function TNeuron.FunctOutpLinearDeriv: Extended;
begin
  Result := FSlope;
end;

function TNeuron.FunctOutpSquashCosDeriv: Extended;
begin
  Result := (FOutputMax - FOutputMin) / 2.0;
  Result := FSlope * (FActivity - FBias) / Result;
  if Result < (-Pi / 2.0) then Result := 0.0
    else if Result > (Pi / 2.0) then Result := 0.0
      else Result := FSlope * Cos(Result);
  Result := 0.1 + Result;
end;

function TNeuron.FunctOutpTanhDeriv: Extended;
begin
  Result := (FOutputMax - FOutputMin) / 2.0;
  Result := FSlope * (FActivity - FBias) / Result;
  Result := Tanh(Result);
  Result := 0.1 + (1.0 - Result * Result) * FSlope;
end;

function TNeuron.FunctOutpUserDeriv: Extended;
begin
  Result := FOutputDeriv;
  if Assigned(FOnUserOutpDerivFunction) then FOnUserOutpDerivFunction(Self, Result);
end;

{ propagation functions }
function TNeuron.FunctPropScalProd: Extended;
var I: Integer;
begin
  Result := 0.0;
  for I := 0 to (FInputNo - 1) do
    Result := Result + FInputs[I] * FWeights[I];
  Result := Result / FInputNo;
end;

function TNeuron.FunctPropSigmaPi: Extended;
begin
  Result := FInputs[0] * FWeights[0];
end;

function TNeuron.FunctPropUser: Extended;
begin
  Result := FNetInput;
  if Assigned(FOnUserPropFunction) then FOnUserPropFunction(Self, Result);
end;
{ end of propagation functions }

function TNeuron.GetRandom: Extended;
begin
  Result := Random * (FParamsMax - FParamsMin);
  Result := Result - FParamsMax;
end;

procedure TNeuron.IncInputNo(Index: Integer);
begin
  if (Index >= 0) and (Index <= FInputNo) then begin
    FInputs.Insert(Index, 0.0);
    FWeights.Insert(Index, GetRandom);
    FWeightsDelta.Insert(Index, 0.0);
    FWeightsDeltaOld.Insert(Index, 0.0);
    FWeightsDelta2Old.Insert(Index, 0.0);
    FAdaptWeights.Insert(Index, FAdaptWeights[0]);
    Inc(FInputNo);
  end;
end;

procedure TNeuron.ResetAdapts(const AdaptConst: Word);
var I: Integer;
begin
  for I := 0 to (FInputNo - 1) do begin
    FAdaptWeights[I] := AdaptConst;
    FWeightsDelta[I] := 0.0;
    FWeightsDeltaOld[I] := 0.0;
    FWeightsDelta2Old[I] := 0.0;
  end;
  FAdaptBias := AdaptConst;
  FBiasDelta := 0.0;
  FBiasDeltaOld := 0.0;
  FBiasDelta2Old := 0.0;
  FAdaptSlope := AdaptConst;
  FSlopeDelta := 0.0;
  FSlopeDeltaOld := 0.0;
  FSlopeDelta2Old := 0.0;
end;

procedure TNeuron.ResetOutputs;
begin
  FActivity := 0.0;
  FActivityAll := 0.0;
  FDelta := 0.0;
  FError := 0.0;
  FNetInput := 0.0;
  FOutput := 0.0;
  FOutputAll := 0.0;
  FOutputDeriv := 0.0;
  FPowerAll := 0.0;
end;

procedure TNeuron.ResetParams;
begin
  SetLearnParams(FLearnParams);
end;

procedure TNeuron.SetActFunction(const Value: TFunctTypeAct);
begin
  if Value <> FActFunction then begin
    FActFunction := Value;
    case FActFunction of
      faUser     : FFunctAct := FunctActUser;
      faLinear   : FFunctAct := FunctActLinear;
      faBSB      : FFunctAct := FunctActBSB;
      faDMA      : FFunctAct := FunctActDMA;
      faHopfield : FFunctAct := FunctActHopfield;
    end;
  end;
end;

procedure TNeuron.SetInputNo(const Value: Integer);
begin
  if (Value <> FInputNo) then begin
    if Value <= 0 then
      raise Exception.Create('Das Neuron mu mindestens einen Eingang besitzen');
    while FInputNo < Value do IncInputNo(FInputNo);
    while FInputNo > Value do DecInputNo(FInputNo - 1);
  end;
end;

procedure TNeuron.SetLearnParams(const Value: TLearnParams);
var I: Integer;
begin
  FLearnParams := Value;
  if (lpWeights in FLearnParams) then for I := 0 to (FInputNo - 1) do FWeights[I] := GetRandom
                                 else for I := 0 to (FInputNo - 1) do FWeights[I] := 0.0;
  if (lpBias in FLearnParams) then FBias := GetRandom
                              else FBias := 0.0;
  if (lpSlope in FLearnParams) then FSlope := GetRandom + 1.0
                               else FSlope := 1.0;
end;

procedure TNeuron.SetOnUserActFunction(Value: TUserProc);
begin
  FOnUserActFunction := Value;
  ActFunction := faUser;
end;

procedure TNeuron.SetOnUserOutpFunction(Value: TUserProc);
begin
  FOnUserOutpFunction := Value;
  OutpFunction := foUser;
end;

procedure TNeuron.SetOnUserPropFunction(Value: TUserProc);
begin
  FOnUserPropFunction := Value;
  PropFunction := fpUser;
end;

procedure TNeuron.SetOutpFunction(const Value: TFunctTypeOutp);
begin
  if Value <> FOutpFunction then begin
    FOutpFunction := Value;
    case FOutpFunction of
      foUser         : begin FFunctOutp := FunctOutpUser;
                             FFunctOutpDeriv := FunctOutpUserDeriv; end;
      foFermi        : begin FFunctOutp := FunctOutpFermi;
                             FFunctOutpDeriv := FunctOutpFermiDeriv; end;
      foTanh         : begin FFunctOutp := FunctOutpTanh;
                             FFunctOutpDeriv := FunctOutpTanhDeriv; end;
      foSquashCos    : begin FFunctOutp := FunctOutpSquashCos;
                             FFunctOutpDeriv := FunctOutpSquashCosDeriv; end;
      foArcTan       : begin FFunctOutp := FunctOutpArcTan;
                             FFunctOutpDeriv := FunctOutpArcTanDeriv; end;
      foLinear       : begin FFunctOutp := FunctOutpLinear;
                             FFunctOutpDeriv := FunctOutpLinearDeriv; end;
      foThreshold    : FFunctOutp := FunctOutpThreshold;
      foThresholdLin : FFunctOutp := FunctOutpThresholdLin;
      foHardLimit    : FFunctOutp := FunctOutpHardLimit;
    end;
  end;
end;

procedure TNeuron.SetParamsRange(Index: Integer; const Value: Extended);
begin
  case Index of
    1: FParamsMax := Value;
    2: FParamsMin := Value;
  end;
  SetLearnParams(FLearnParams);
end;

procedure TNeuron.SetPropFunction(const Value: TFunctTypeProp);
begin
  if Value <> FPropFunction then begin
    FPropFunction := Value;
    case FPropFunction of
      fpUser     : FFunctProp := FunctPropUser;
      fpScalProd : FFunctProp := FunctPropScalProd;
      fpSigmaPi  : FFunctProp := FunctPropSigmaPi;
    end;
  end;
end;


{ methods of TLayer }
constructor TLayer.Create(cNeuronNo, cInputNo: Word);
var I: Integer; Neuron: TNeuron;
begin
  if (cNeuronNo = 0) or (cInputNo = 0) then
    raise Exception.Create('Kann Schicht nicht erstellen.' + #13#10 +
            'Die Schicht mu mindestens ein Neuron mit einem Eingang enthalten');
  inherited Create;
  List := TList.Create;
  FNeuronNo := cNeuronNo;
  List.Capacity := FNeuronNo;
  for I := 0 to (FNeuronNo - 1) do begin
    Neuron := TNeuron.Create(cInputNo);
    List.Add(Neuron);
  end;
end;

destructor TLayer.Destroy;
var I: Integer;
begin
  for I := 0 to (FNeuronNo - 1) do
    Items[I].Free;
  List.Free;
  inherited Destroy;
end;

procedure TLayer.Assign(Source: TPersistent);
var I: Integer;
begin
  if Source is TLayer then
    with Source as TLayer do
      for I := 0 to (Self.NeuronNo - 1) do
        Self.Items[I].Assign(Items[0])
  else inherited Assign(Source);
end;

procedure TLayer.DeleteNeuron(Index: Integer);
var I: Integer;
begin
  if (Index >= 0) and (Index < FNeuronNo) then begin
    if FNeuronNo = 1 then
      raise Exception.Create('Kann Neuron nicht lschen.' + #13#10 +
                             'Jede Schicht mu mindestens ein Neuron enthalten');
    Items[Index].Free;
    List.Delete(Index);
    Dec(FNeuronNo);
    if FNextLayer <> nil then    { if not output layer then.. }
      with FNextLayer do         { ..update next layer }
        for I := 0 to (NeuronNo - 1) do
          Items[I].DecInputNo(Index);
  end;
end;

procedure TLayer.FeedForward;
var I, J: Integer;
begin
  if FPrevLayer = nil then
    raise Exception.Create('Keine Vorgngerschicht gefunden');
  for I := 0 to (FNeuronNo - 1) do
    with Items[I] do begin
      for J := 0 to (InputNo - 1) do
        Inputs[J] := FPrevLayer[J].Output;
      CalcState;
    end;
end;

function TLayer.Get(Index: Integer): TNeuron;
begin
  Result := TNeuron(List.Items[Index]);
end;

function TLayer.GetActFunction: TFunctTypeAct;
begin
  Result := Items[0].ActFunction;
end;

function TLayer.GetBias: Extended;
begin
  Result := Items[0].Bias;
end;

function TLayer.GetOnUserActFunction: TUserProc;
begin
  Result := Items[0].OnUserActFunction;
end;

function TLayer.GetOnUserOutpDerivFunction: TUserProc;
begin
  Result := Items[0].OnUserOutpDerivFunction;
end;

function TLayer.GetOnUserOutpFunction: TUserProc;
begin
  Result := Items[0].OnUserOutpFunction;
end;

function TLayer.GetOnUserPropFunction: TUserProc;
begin
  Result := Items[0].OnUserPropFunction;
end;

function TLayer.GetOutpFunction: TFunctTypeOutp;
begin
  Result := Items[0].OutpFunction;
end;

function TLayer.GetOutputMax: Extended;
begin
  Result := Items[0].OutputMax;
end;

function TLayer.GetOutputMin: Extended;
begin
  Result := Items[0].OutputMin;
end;

function TLayer.GetPropFunction: TFunctTypeProp;
begin
  Result := Items[0].PropFunction;
end;

function TLayer.GetScale: Extended;
begin
  Result := Items[0].Scale;
end;

function TLayer.GetSlope: Extended;
begin
  Result := Items[0].Slope;
end;

procedure TLayer.InsertNeuron(Index: Integer);
var I: Integer; Neuron: TNeuron;
begin
  if (Index >= 0) and (Index <= FNeuronNo) then begin
    Neuron := TNeuron.Create(Items[0].InputNo);
    Neuron.Assign(Items[0]);     { use first neuron of layer to expand, to initialize new neuron }
    List.Insert(Index, Neuron);
    Inc(FNeuronNo);
    if FNextLayer <> nil then    { if not output layer then.. }
      with FNextLayer do         { ..update next layer }
        for I := 0 to (NeuronNo - 1) do
          Items[I].IncInputNo(Index);
  end;
end;

procedure TLayer.LoadFromStream(Stream: TStream);
var IntVal: Integer; DoubleVal: Extended;
begin
  Stream.Read(IntVal, SizeOf(IntVal));
  NeuronNo := IntVal;
  Stream.Read(IntVal, SizeOf(IntVal));
  ActFunction := TFunctTypeAct(IntVal);
  Stream.Read(IntVal, SizeOf(IntVal));
  OutpFunction := TFunctTypeOutp(IntVal);
  Stream.Read(IntVal, SizeOf(IntVal));
  PropFunction := TFunctTypeProp(IntVal);
  Stream.Read(DoubleVal, SizeOf(DoubleVal));
  Scale := DoubleVal;
  if not(lpBias in Items[0].LearnParams) then begin
    Stream.Read(DoubleVal, SizeOf(DoubleVal));
    Bias := DoubleVal;
  end;
  if not(lpSlope in Items[0].LearnParams) then begin
    Stream.Read(DoubleVal, SizeOf(DoubleVal));
    Slope := DoubleVal;
  end;
  Stream.Read(DoubleVal, SizeOf(DoubleVal));
  OutputMax := DoubleVal;
  Stream.Read(DoubleVal, SizeOf(DoubleVal));
  OutputMin := DoubleVal;
end;

procedure TLayer.SaveToStream(Stream: TStream);
var IntVal: Integer; DoubleVal: Extended;
begin
  Stream.Write(FNeuronNo, SizeOf(FNeuronNo));
  IntVal := Ord(ActFunction);
  Stream.Write(IntVal, SizeOf(IntVal));
  IntVal := Ord(OutpFunction);
  Stream.Write(IntVal, SizeOf(IntVal));
  IntVal := Ord(PropFunction);
  Stream.Write(IntVal, SizeOf(IntVal));
  DoubleVal := Scale;
  Stream.Write(DoubleVal, SizeOf(DoubleVal));
  if not(lpBias in Items[0].LearnParams) then begin
    DoubleVal := Bias;
    Stream.Write(DoubleVal, SizeOf(DoubleVal));
  end;
  if not(lpSlope in Items[0].LearnParams) then begin
    DoubleVal := Slope;
    Stream.Write(DoubleVal, SizeOf(DoubleVal));
  end;
  DoubleVal := OutputMax;
  Stream.Write(DoubleVal, SizeOf(DoubleVal));
  DoubleVal := OutputMin;
  Stream.Write(DoubleVal, SizeOf(DoubleVal));
end;

procedure TLayer.SetActFunction(const Value: TFunctTypeAct);
var I: Integer;
begin
  if Value <> Items[0].Actfunction then
    for I := 0 to (FNeuronNo - 1) do
      Items[I].ActFunction := Value;
end;

procedure TLayer.SetBias(const Value: Extended);
var I: Integer;
begin
  if Value <> Items[0].Bias then begin
    if (lpBias in Items[0].LearnParams) then
      raise Exception.Create('Kann ''Bias'' nicht setzen,' + #13#10 +
                        'whrend ''LearnParams'' das Element [lpBias] enthlt');
    for I := 0 to (FNeuronNo - 1) do
      Items[I].Bias := Value;
   end;
end;

procedure TLayer.SetNeuronNo(const Value: Integer);
begin
  if Value <> FNeuronNo then begin
    if Value <= 0 then
      raise Exception.Create('Jede Schicht mu mindestens ein Neuron enthalten');
    while FNeuronNo < Value do
      InsertNeuron(FNeuronNo);
    while FNeuronNo > Value do
      DeleteNeuron(FNeuronNo - 1);
  end;
end;

procedure TLayer.SetOnUserActFunction(Value: TUserProc);
var I: Integer;
begin
  for I := 0 to (FNeuronNo - 1) do
    Items[I].OnUserActFunction := Value;
end;

procedure TLayer.SetOnUserOutpDerivFunction(Value: TUserProc);
var I: Integer;
begin
  for I := 0 to (FNeuronNo - 1) do
    Items[I].OnUserOutpDerivFunction := Value;
end;

procedure TLayer.SetOnUserOutpFunction(Value: TUserProc);
var I: Integer;
begin
  for I := 0 to (FNeuronNo - 1) do
    Items[I].OnUserOutpFunction := Value;
end;

procedure TLayer.SetOnUserPropFunction(Value: TUserProc);
var I: Integer;
begin
  for I := 0 to (FNeuronNo - 1) do
    Items[I].OnUserPropFunction := Value;
end;

procedure TLayer.SetOutpFunction(const Value: TFunctTypeOutp);
var I: Integer;
begin
  if Value <> Items[0].OutpFunction then
    for I := 0 to (FNeuronNo - 1) do
      Items[I].OutpFunction := Value;
end;

procedure TLayer.SetOutputMax(const Value: Extended);
var I: Integer;
begin
  if Value <> Items[0].OutputMax then begin
    if Value <= Items[0].OutputMin then
      raise Exception.Create('''OutputMax'' mu ein Gleitkommawert' + #13#10 +
                             'und grer als ''OutputMin'' sein');
    for I := 0 to (FNeuronNo - 1) do
      Items[I].OutputMax := Value;
  end;
end;

procedure TLayer.SetOutputMin(const Value: Extended);
var I: Integer;
begin
  if Value <> Items[0].OutputMin then begin
    if Value >= Items[0].OutputMax then
      raise Exception.Create('''OutputMin'' mu ein Gleitkommawert' + #13#10 +
                             'und kleiner als ''OutputMax'' sein');
    for I := 0 to (FNeuronNo - 1) do
      Items[I].OutputMin := Value;
  end;
end;

procedure TLayer.SetPropFunction(const Value: TFunctTypeProp);
var I: Integer;
begin
  if Value <> Items[0].PropFunction then
    for I := 0 to (FNeuronNo - 1) do
      Items[I].PropFunction := Value;
end;

procedure TLayer.SetScale(const Value: Extended);
var I: Integer;
begin
  if Value <> Items[0].Scale then
    for I := 0 to (FNeuronNo - 1) do
      Items[I].Scale := Value;
end;

procedure TLayer.SetSlope(const Value: Extended);
var I: Integer;
begin
  if Value <> Items[0].Slope then begin
    if (lpSlope in Items[0].LearnParams) then
      raise Exception.Create('Kann ''Slope'' nicht setzen,' + #13#10 +
                        'whrend ''LearnParams'' das Element [lpSlope] enthlt');
    for I := 0 to (FNeuronNo - 1) do
      Items[I].Slope := Value;
   end;
end;


{ methods of TBackPropNet }
constructor TBackPropNet.Create(AOwner: TComponent);
var NewLayer: TLayer;
begin
  inherited Create(AOwner);
  FInputNo := 1;
  FLayerNo := 1;
  FAdaptConstFactor := 1;
  FLearnParams := [lpWeights, lpBias];
  FLearnRate := 0.7;
  FLearnRateAdapt := 0.7;
  FLearnRateDec := 1.0;
  FLearnRateInc := 1.0;
  FMomentum := 0.9;
  FCycleLimit := 640;
  FErrorLimit := 0.00001;
  FInterval := 1;
  FParamsMax := 0.9;
  FParamsMin := -0.9;
  Randomize;
  List := TList.Create;
  NewLayer := TLayer.Create(1, 1);
  NewLayer.Net := Self;
  NewLayer.Selected := True;
  List.Add(NewLayer);
  FPatternSet := TSet.Create(0, 0, 0);
  FRecallSet := TSet.Create(0, 0, 0);
end;

destructor TBackPropNet.Destroy;
var I: Integer;
begin
  for I := 0 to (FLayerNo - 1) do
    Items[I].Free;
  List.Free;
  FPatternSet.Free;
  FRecallSet.Free;
  inherited Destroy;
end;

procedure TBackPropNet.BackProp;
var I, J, K: Integer;
begin
  with Items[FLayerNo - 1] do            { estimation of delta value of output layer first }
    for I := 0 to (NeuronNo - 1) do
      with Items[I] do
        Delta := (Target - Output) * OutputDeriv;
  if FLayerNo > 1 then                   { if exist hidden layers }
    for I := (FlayerNo - 2) downto 0 do  { estimation of delta values of all hidden layers.. }
      with Items[I] do                   { ..in opposite order afterwards }
        for J := 0 to (NeuronNo - 1) do
          with Items[J] do begin
            Delta := 0.0;
            for K := 0 to (NextLayer.NeuronNo - 1) do
              Delta := Delta + NextLayer[K].Delta * NextLayer[K].Weights[J];
            Delta := Delta * OutputDeriv;
          end;
  case FLearnMethod of
    lmEpochal: begin   { evaluates gradients of all layers }
                 for I := 0 to (FLayerNo - 1) do
                   with Items[I] do
                     for J := 0 to (NeuronNo - 1) do
                       with Items[J] do begin
                         if (lpWeights in FLearnParams) then
                           for K := 0 to (InputNo - 1) do
                             WeightsDelta[K] := WeightsDelta[K] + FLearnRateAdapt * Delta * Inputs[K];
                         if (lpBias in FLearnParams) then
                           BiasDelta := BiasDelta - FLearnRateAdapt * Delta;
                         if (lpSlope in FLearnParams) then
                           SlopeDelta := SlopeDelta + FLearnRateAdapt * Delta * NetInput;
                       end;
               end;
    lmDirect: begin    { evaluates gradients of all layers and updates Weights, Bias, Slope }
                for I := 0 to (FLayerNo - 1) do
                  with Items[I] do
                    for J := 0 to (NeuronNo - 1) do
                      with Items[J] do begin
                         if (lpWeights in FLearnParams) then
                           for K := 0 to (InputNo - 1) do begin
                             WeightsDelta[K] := FLearnRateAdapt * Delta * Inputs[K];
                             Weights[K] := Weights[K] + WeightsDelta[K];
                           end;
                         if (lpBias in FLearnParams) then begin
                           BiasDelta := -FLearnRateAdapt * Delta;
                           Bias := Bias + BiasDelta;
                         end;
                         if (lpSlope in FLearnParams) then begin
                           SlopeDelta := FLearnRateAdapt * Delta * NetInput;
                           Slope := Slope + SlopeDelta;
                         end;
                       end;
              end;
  end;
end;

procedure TBackPropNet.DefineProperties(Filer: TFiler);

  function DoPatternSet: Boolean;
  begin
    Result := False;
    if FPatternSet.PatternNo <> 0 then
      Result := True;
  end;

  function DoRecallSet: Boolean;
  begin
    Result := False;
    if FRecallSet.PatternNo <> 0 then
      Result := True;
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Layers', LoadLayersFromStream, SaveLayersToStream, True);
  Filer.DefineBinaryProperty('Pattern', FPatternSet.LoadFromStream, FPatternSet.SaveToStream, DoPatternSet);
  Filer.DefineBinaryProperty('Recall', FRecallSet.LoadFromStream, FRecallSet.SaveToStream, DoRecallSet);
end;

procedure TBackPropNet.DeleteLayer(Index: Integer);
var I: Integer;
begin
  if (Index >= 0) and (Index < FLayerNo) then begin
    if FLayerNo = 1 then
      raise Exception.Create('Kann Schicht nicht lschen.' + #13#10 +
                             'Das Netz mu mindestens eine Schicht enthalten');
    Items[Index].Free;
    List.Delete(Index);
    if Index = 0 then
      with Items[0] do begin
        PrevLayer := nil;
        for I := 0 to (NeuronNo - 1) do
          Items[I].InputNo := FInputNo;
      end;
    if (Index > 0) and (Index < (FLayerNo - 1)) then begin
      Items[Index].PrevLayer := Items[Index - 1];
      Items[Index - 1].NextLayer := Items[Index];
      with Items[Index] do          { update next layer }
        for I := 0 to (NeuronNo - 1) do
          Items[I].InputNo := PrevLayer.NeuronNo;
    end;
    if Index = (FLayerNo - 1) then
      Items[Index - 1].NextLayer := nil;
    Dec(FLayerNo);
  end;
end;

procedure TBackPropNet.DeleteNeuron(Index, NeuronIndex: Integer);
begin
  if (Index >= 0) and (Index < FLayerNo) then
    Items[Index].DeleteNeuron(NeuronIndex);
end;

procedure TBackPropNet.FeedForward;
var I: Integer;
begin
  if FLayerNo > 1 then
    for I := 1 to (FLayerNo - 1) do
      Items[I].FeedForward;
end;

function TBackPropNet.Get(Index: Integer): TLayer;
begin
  Result := TLayer(List.Items[Index]);
end;

function TBackPropNet.GetLayer: TLayer;
begin
  Result := Items[0];
end;

{function TBackPropNet.GetMaxGradient: Extended;
var I, J: Integer;
begin
  Result := 0.0;
  with Items[FLayerNo - 1] do
    for I := 0 to (NeuronNo - 1) do
      with Items[I] do begin
        if (lpWeights in FLearnParams) then
          for J := 0 to (InputNo - 1) do
            if abs(WeightsDelta[J]) > Result then
              Result := abs(WeightsDelta[J]);
        if (lpBias in FLearnParams) and (abs(BiasDelta) > Result) then
          Result := abs(BiasDelta);
        if (lpSlope in FLearnParams) and (abs(SlopeDelta) > Result) then
          Result := abs(SlopeDelta);
      end;
end;}

function TBackPropNet.GetPatternError: Extended;
var I: Integer;
begin
  Result := 0.0;
  with Items[FLayerNo - 1] do begin
    for I := 0 to (NeuronNo - 1) do
      with Items[I] do
        Result := Result + sqr(Error);
    Result := Result / NeuronNo;
  end;
end;

procedure TBackPropNet.InsertLayer(Index, Size: Integer);
var I: Integer; NewLayer: TLayer;
begin
  if Size < 1 then
    raise Exception.Create('Kann Schicht nicht einfgen.' + #13#10 +
                           'Jede Schicht mu mindestens ein Neuron enthalten');
  if (Index >= 0) and (Index <= FLayerNo) then begin
    if Index = 0 then NewLayer := TLayer.Create(Size, FInputNo)
                 else NewLayer := TLayer.Create(Size, Items[Index - 1].NeuronNo);
    NewLayer.Net := Self;
    NewLayer.Assign(Items[0]);             { use first layer to initialize new layer}
    List.Insert(Index, NewLayer);
    if Index > 0 then begin
      Items[Index].PrevLayer := Items[Index - 1];
      Items[Index - 1].NextLayer := Items[Index];
    end;
    if Index < FLayerNo then begin
      Items[Index].NextLayer := Items[Index + 1];
      Items[Index + 1].PrevLayer := Items[Index];
      with Items[Index].NextLayer do    { update next layer }
        for I := 0 to (NeuronNo - 1) do
          Items[I].InputNo := Size;
    end;
    Inc(FLayerNo);
    end;
end;

procedure TBackPropNet.InsertNeuron(Index, NeuronIndex: Integer);
begin
  if (Index >= 0) and (Index < FLayerNo) then
    Items[Index].InsertNeuron(NeuronIndex);
end;

procedure TBackPropNet.Learn;
var I, CycleCount: Integer;
const Patterns: set of Byte = [];

  procedure ResetDeltas;
  var I, J, K: Integer;
  begin
    for I := 0 to (FLayerNo - 1) do
    with Items[I] do
      for J := 0 to (NeuronNo - 1) do
      with Items[J] do begin
        if (lpWeights in FLearnParams) then
          for K := 0 to (InputNo - 1) do WeightsDelta[K] := 0.0;
        if (lpBias in FLearnParams) then BiasDelta := 0.0;
        if (lpSlope in FLearnParams) then SlopeDelta := 0.0;
      end;
  end;

begin
  if FPatternSet.PatternNo = 0 then
    raise Exception.Create('Das Musterset enthlt kein Muster');
  if FPatternSet.InputNo <> FInputNo then
    raise Exception.Create('Anzahl der Netzeingnge und' + #13#10 +
                           'Gre des Eingangsvektors stimmen nicht berein');
  if FPatternSet.OutputNo <> Items[FLayerNo - 1].NeuronNo then
    raise Exception.Create('Anzahl der Netzausgnge und' + #13#10 +
                           'Gre des Ausgangsvektors stimmen nicht berein');
  CycleCount := 0;
  FCycles := 0;
  while FCycles < FCycleLimit do begin
    Application.ProcessMessages;
    if FBreak then begin
      FBreak := False;
      if Assigned(FOnExit) then FOnExit(Self, exBreak);
      Exit;
    end;
    ResetOutputs;
    case FLearnMethod of
      lmEpochal: begin
                   ResetDeltas;
                   for I := 0 to (FPatternSet.PatternNo - 1) do begin
                     SetTargets(FPatternSet[I].Outputs);
                     SetInputs(FPatternSet[I].Inputs);
                     FeedForward;
                     FGlobalError := FGlobalError + GetPatternError;
                     BackProp;
                   end;
                   SetEpochWeights;
                 end;
      lmDirect: begin
                  for I := 0 to (FPatternSet.PatternNo - 1) do
                    Include(Patterns, I);
                  while Patterns <> [] do begin
                    repeat         { selection of pattern by chance }
                      I := Random(FPatternSet.PatternNo);
                    until (I in Patterns);
                    SetTargets(FPatternSet[I].Outputs);
                    SetInputs(FPatternSet[I].Inputs);
                    FeedForward;
                    FGlobalError := FGlobalError + GetPatternError;
                    BackProp;
                    Exclude(Patterns, I);
                  end;
                end;
    end;
    FGlobalError := FGlobalError / FPatternSet.PatternNo;
    Inc(FCycles);
    Inc(CycleCount);
    if CycleCount = FInterval then begin
      if Assigned(FOnInterval) then FOnInterval(Self);
      CycleCount := 0;
    end;
    if FGlobalError <= FErrorLimit then begin
      if Assigned(FOnExit) then FOnExit(Self, exError);
      Exit;
    end;
    if (FLearnRateDec < 1) or (FLearnRateInc > 1) then begin
      if ((FGlobalError2Old <= FGlobalErrorOld) and (FGlobalErrorOld <= FGlobalError)) or
         ((FGlobalError2Old > FGlobalErrorOld) and (FGlobalErrorOld > FGlobalError))
        then begin
          FLearnRateAdapt := FLearnRateAdapt * FLearnRateInc;
          if FLearnRateAdapt > 10 * FLearnRate then FLearnRateAdapt := 10 * FLearnRate;
        end
        else begin
          FLearnRateAdapt := FLearnRateAdapt * FLearnRateDec;
          if FLearnRateAdapt < 0.1 * FLearnRate then FLearnRateAdapt := 0.1 * FLearnRate;
        end;
      FGlobalError2Old := FGlobalErrorOld;
      FGlobalErrorOld := FGlobalError;
    end;
  end;
  if Assigned(FOnExit) then FOnExit(Self, exCycle);
end;

procedure TBackPropNet.LoadConfigFromStream(Stream: TStream);
var
  I, IntVal: Integer;
  DoubleVal: Extended;
begin
  Stream.Read(IntVal, SizeOf(IntVal));
  LayerNo := IntVal;
  Stream.Read(IntVal, SizeOf(IntVal));
  InputNo := IntVal;
  for I := 0 to (FLayerNo - 1) do
    Stream.Read(IntVal, SizeOf(IntVal));
  Stream.Read(IntVal, SizeOf(IntVal));
  SetOrdProp(Self, GetPropInfo(PTypeInfo(TBackPropNet.ClassInfo), 'LearnParams'), IntVal);
  LoadLayersFromStream(Stream);
  Stream.Read(IntVal, SizeOf(IntVal));
  AdaptConst := IntVal;
  Stream.Read(FAdaptConstFactor, SizeOf(FAdaptConstFactor));
  Stream.Read(FAdaptConstStep, SizeOf(FAdaptConstStep));
  Stream.Read(FLearnRate, SizeOf(FLearnRate));
  Stream.Read(FLearnRateDec, SizeOf(FLearnRateDec));
  Stream.Read(FLearnRateInc, SizeOf(FLearnRateInc));
  Stream.Read(FMomentum, SizeOf(FMomentum));
  Stream.Read(IntVal, SizeOf(IntVal));
  FLearnMethod := TLearnMethod(IntVal);
  Stream.Read(DoubleVal, SizeOf(DoubleVal));
  ParamsMax := DoubleVal;
  Stream.Read(DoubleVal, SizeOf(DoubleVal));
  ParamsMin := DoubleVal;
  Stream.Read(FCycleLimit, SizeOf(FCycleLimit));
  Stream.Read(FErrorLimit, SizeOf(FErrorLimit));
  Stream.Read(FInterval, SizeOf(FInterval));
end;

procedure TBackPropNet.LoadFromFile(const FileName: string);
var
  Ext: string;
  TestSignature: string[9];
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    Stream.Read(TestSignature[1], 9);
    TestSignature[0] := #9;
    if AnsiCompareText(FileSignature, TestSignature) <> 0 then
      raise Exception.Create('Datei ''' + FileName + ''' ist keine gltige Netzdatei!');
    Ext := LowerCase(ExtractFileExt(FileName));
    if Ext = '.cfg' then LoadConfigFromStream(Stream)
      else if Ext = '.lrn' then LoadParamsFromStream(Stream)
        else if Ext = '.net' then LoadNetFromStream(Stream)
          else if Ext = '.ptn' then FPatternSet.LoadFromStream(Stream)
            else if Ext = '.rcl' then FRecallSet.LoadFromStream(Stream)
              else begin
                Stream.Free;
                Stream := TMemoryStream.Create;
                TMemoryStream(Stream).SetSize(4096);
                FileDlg.ConvertAscFileToStream(FileName, Stream);
                FPatternSet.LoadFromStream(Stream);
              end;
  finally
    Stream.Free;
  end;
end;

procedure TBackPropNet.LoadLayersFromStream(Stream: TStream);
var I: Integer;
begin
  for I := 0 to (FLayerNo - 1) do
    Items[I].LoadFromStream(Stream);
end;

procedure TBackPropNet.LoadNetFromStream(Stream: TStream);
var I, J, K: Integer; DoubleVal: Extended;
begin
  LoadConfigFromStream(Stream);
  FPatternSet.LoadFromStream(Stream);
  FRecallSet.LoadFromStream(Stream);
  { load learned params }
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
      for J := 0 to (NeuronNo - 1) do
        with Items[J] do begin
          for K := 0 to (InputNo - 1) do begin
            Stream.Read(DoubleVal, SizeOf(DoubleVal));
            Weights[K] := DoubleVal;
          end;
          Stream.Read(DoubleVal, SizeOf(DoubleVal));
          Bias := DoubleVal;
          Stream.Read(DoubleVal, SizeOf(DoubleVal));
          Slope := DoubleVal;
        end;
end;

procedure TBackPropNet.LoadParamsFromStream(Stream: TStream);
var
  I, J, K: Integer;
  DoubleVal: Extended;
begin
  Stream.Read(I, SizeOf(I));  { read LayerNo }
  if I <> FLayerNo then
    raise Exception.Create('Netzkonfiguration und Dateidaten (Schichtanzahl) stimmen nicht berein');
  Stream.Read(I, SizeOf(I));  { read InputNo }
  if I <> FInputNo then
    raise Exception.Create('Netzkonfiguration und Dateidaten (Anzahl der Eingnge) stimmen nicht berein');
  for I := 0 to (FLayerNo - 1) do begin  { read NeuronNo's of all layers }
    Stream.Read(J, SizeOf(J));
    if J <> Items[I].NeuronNo then
      raise Exception.Create('Netzkonfiguration und Dateidaten (Neuronenanzahl) stimmen nicht berein');
  end;
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
      for J := 0 to (NeuronNo - 1) do
        with Items[J] do begin
          for K := 0 to (InputNo - 1) do begin
            Stream.Read(DoubleVal, SizeOf(DoubleVal));
            Weights[K] := DoubleVal;
          end;
          Stream.Read(DoubleVal, SizeOf(DoubleVal));
          Bias := DoubleVal;
          Stream.Read(DoubleVal, SizeOf(DoubleVal));
          Slope := DoubleVal;
        end;
end;

procedure TBackPropNet.Recall;
var I: Integer;
begin
  if FRecallSet.PatternNo = 0 then
    raise Exception.Create('Das Musterset enthlt kein Muster');
  if FRecallSet.InputNo <> FInputNo then
    raise Exception.Create('Anzahl der Netzeingnge und' + #13#10 +
                           'Gre des Eingangsvektors stimmen nicht berein');
  if FRecallSet.OutputNo <> Items[FLayerNo - 1].NeuronNo then
    raise Exception.Create('Anzahl der Netzausgnge und' + #13#10 +
                           'Gre des Ausgangsvektors stimmen nicht berein');
  for I := 0 to (FRecallSet.PatternNo - 1) do begin
    Application.ProcessMessages;
    if FBreak then begin
      FBreak := False;
      if Assigned(FOnExit) then FOnExit(Self, exBreak);
      Exit;
    end;
    SetInputs(FRecallSet[I].Inputs);
    FeedForward;
    SetOutputs(FRecallSet[I].Outputs);
    if Assigned(FOnInterval) then FOnInterval(Self);
  end;
  if Assigned(FOnExit) then FOnExit(Self, exCycle);
end;

procedure TBackPropNet.ResetAdapts;
var I, J: Integer;
begin
  FLearnRateAdapt := FLearnRate;
  FGlobalErrorOld := 0.0;
  FGlobalError2Old := 0.0;
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
      for J := 0 to (NeuronNo - 1) do
        Items[J].ResetAdapts(FAdaptConst);
end;

procedure TBackPropNet.ResetAll;
begin
  FCycles := 0;
  ResetAdapts;
  ResetOutputs;
  ResetParams;
end;

procedure TBackPropNet.ResetOutputs;
var I, J: Integer;
begin
  FGlobalError := 0.0;
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
      for J := 0 to (NeuronNo - 1) do
        Items[J].ResetOutputs;
end;

procedure TBackPropNet.ResetParams;
var I, J: Integer;
begin
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
      for J := 0 to (NeuronNo - 1) do
        Items[J].ResetParams;
end;

procedure TBackPropNet.SaveConfigToStream(Stream: TStream);
var I, IntVal: Integer;
begin
  Stream.Write(FLayerNo, SizeOf(FLayerNo));
  Stream.Write(FInputNo, SizeOf(FInputNo));
  for I := 0 to (FLayerNo - 1) do begin
    IntVal := Items[I].NeuronNo;
    Stream.Write(IntVal, SizeOf(IntVal));
  end;
  IntVal := GetOrdProp(Self, GetPropInfo(PTypeInfo(TBackPropNet.ClassInfo), 'LearnParams'));
  Stream.Write(IntVal, SizeOf(IntVal));
  SaveLayersToStream(Stream);
  Stream.Write(FAdaptConst, SizeOf(FAdaptConst));
  Stream.Write(FAdaptConstFactor, SizeOf(FAdaptConstFactor));
  Stream.Write(FAdaptConstStep, SizeOf(FAdaptConstStep));
  Stream.Write(FLearnRate, SizeOf(FLearnRate));
  Stream.Write(FLearnRateDec, SizeOf(FLearnRateDec));
  Stream.Write(FLearnRateInc, SizeOf(FLearnRateInc));
  Stream.Write(FMomentum, SizeOf(FMomentum));
  IntVal := Ord(FLearnMethod);
  Stream.Write(IntVal, SizeOf(IntVal));
  Stream.Write(FParamsMax, SizeOf(FParamsMax));
  Stream.Write(FParamsMin, SizeOf(FParamsMin));
  Stream.Write(FCycleLimit, SizeOf(FCycleLimit));
  Stream.Write(FErrorLimit, SizeOf(FErrorLimit));
  Stream.Write(FInterval, SizeOf(FInterval));
end;

procedure TBackPropNet.SaveLayersToStream(Stream: TStream);
var I: Integer;
begin
  for I := 0 to (FLayerNo - 1) do
    Items[I].SaveToStream(Stream);
end;

procedure TBackPropNet.SaveNetToStream(Stream: TStream);
var I, J, K: Integer; DoubleVal: Extended;
begin
  SaveConfigToStream(Stream);
  FPatternSet.SaveToStream(Stream);
  FRecallSet.SaveToStream(Stream);
  { save learned params }
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
    for J := 0 to (NeuronNo - 1) do
      with Items[J] do begin
        for K := 0 to (InputNo - 1) do begin
          DoubleVal := Weights[K];
          Stream.Write(DoubleVal, SizeOf(DoubleVal));
        end;
        DoubleVal := Bias;
        Stream.Write(DoubleVal, SizeOf(DoubleVal));
        DoubleVal := Slope;
        Stream.Write(DoubleVal, SizeOf(DoubleVal));
      end;
end;

procedure TBackPropNet.SaveParamsToStream(Stream: TStream);
var I, J, K: Integer; DoubleVal: Extended;
begin
  Stream.Write(FLayerNo, SizeOf(FLayerNo));
  Stream.Write(FInputNo, SizeOf(FInputNo));
  for I := 0 to (FLayerNo - 1) do begin
    J := Items[I].NeuronNo;
    Stream.Write(J, SizeOf(J));
  end;
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
    for J := 0 to (NeuronNo - 1) do
      with Items[J] do begin
        for K := 0 to (InputNo - 1) do begin
          DoubleVal := Weights[K];
          Stream.Write(DoubleVal, SizeOf(DoubleVal));
        end;
        DoubleVal := Bias;
        Stream.Write(DoubleVal, SizeOf(DoubleVal));
        DoubleVal := Slope;
        Stream.Write(DoubleVal, SizeOf(DoubleVal));
      end;
end;

procedure TBackPropNet.SaveToFile(const FileName: string);
var Ext: string; Stream: TStream;
begin
  Ext := LowerCase(ExtractFileExt(FileName));
  if (Ext = '.cfg') or (Ext = '.lrn') or (Ext = '.net') or
     (Ext = '.ptn') or (Ext = '.rcl') then begin
    Stream := TFileStream.Create(FileName, fmCreate);
    try
      Stream.Write(FileSignature[1], 9);
      if Ext = '.cfg' then SaveConfigToStream(Stream)
        else if Ext = '.lrn' then SaveParamsToStream(Stream)
          else if Ext = '.net' then SaveNetToStream(Stream)
            else if Ext = '.ptn' then FPatternSet.SaveToStream(Stream)
              else FRecallSet.SaveToStream(Stream);
    finally
      Stream.Free;
    end;
  end
  else raise Exception.Create('Falsche Dateierweiterung. Es wurden keine Daten gespeichert!');
end;

procedure TBackPropNet.SetAdaptConst(const Value: Integer);
begin
  if Value <> FAdaptConst then begin
    if Value < 0 then
      raise Exception.Create('''AdaptConst'' mu ein positiver Integerwert sein');
    FAdaptConst := Value;
    ResetAdapts;
    if FAdaptConst <> 0 then FMomentum := 0.0;
  end;
end;

procedure TBackPropNet.SetAdaptConstFactor(const Value: Extended);
begin
  if Value <> FAdaptConstFactor then begin
    if Value < 1 then
      raise Exception.Create('''AdaptConstFactor'' mu ein Gleitkommawert grer oder gleich Eins sein');
    FAdaptConstFactor := Value;
    FAdaptConstStep := 0.0;
  end;
end;

procedure TBackPropNet.SetAdaptConstStep(const Value: Extended);
begin
  if Value <> FAdaptConstStep then begin
    if Value < 0 then
      raise Exception.Create('''AdaptConstStep'' mu ein Gleitkommawert grer oder gleich Null sein');
    FAdaptConstStep := Value;
    FAdaptConstFactor := 1.0;
  end;
end;

procedure TBackPropNet.SetEpochWeights;
var I, J, K: Integer;
begin
  for I := 0 to (FLayerNo - 1) do
    with Items[I] do
      for J := 0 to (NeuronNo - 1) do
        with Items[J] do begin
          if (lpWeights in FLearnParams) then { learning weights? }
            for K := 0 to (InputNo - 1) do begin
              if FAdaptConst > 0 then begin { using adaptive algorithm? }
                WeightsDelta[K] := WeightsDeltaOld[K] + 1.0 / AdaptWeights[K] * (WeightsDelta[K] - WeightsDeltaOld[K]);
                if (FAdaptConstFactor  > 1) or (FAdaptConstStep > 0) then begin { using variable adaptation constant? }
                  if (((WeightsDelta2Old[K] <= WeightsDeltaOld[K]) and (WeightsDeltaOld[K] <= WeightsDelta[K])) or
                      ((WeightsDelta2Old[K] > WeightsDeltaOld[K]) and (WeightsDeltaOld[K] > WeightsDelta[K])))
                  then begin
                    if FAdaptConstStep > 0 then AdaptWeights[K] := AdaptWeights[K] - FAdaptConstStep
                                           else AdaptWeights[K] := AdaptWeights[K] / FAdaptConstFactor;
                    if AdaptWeights[K] < 2.0 then AdaptWeights[K] := 2.0;
                  end
                  else begin
                    if FAdaptConstStep > 0 then AdaptWeights[K] := AdaptWeights[K] + FAdaptConstStep
                                           else AdaptWeights[K] := AdaptWeights[K] * FAdaptConstFactor;
                    if AdaptWeights[K] > FAdaptConst then AdaptWeights[K] := FAdaptConst;
                  end;
                  WeightsDelta2Old[K] := WeightsDeltaOld[K];
                end;
              end;
              Weights[K] := Weights[K] + WeightsDelta[K] + FMomentum * WeightsDeltaOld[K];
              WeightsDeltaOld[K] := WeightsDelta[K];
            end;
          if (lpBias in FLearnParams) then begin { learning bias? }
            if FAdaptConst > 1 then begin { using adaptive algorithm? }
              BiasDelta := BiasDeltaOld + 1.0 / AdaptBias * (BiasDelta - BiasDeltaOld);
              if (FAdaptConstFactor > 1) or (FAdaptConstStep > 0) then begin { using variable adaptation constant? }
                if (((BiasDelta2Old <= BiasDeltaOld) and (BiasDeltaOld <= BiasDelta)) or
                    ((BiasDelta2Old > BiasDeltaOld) and (BiasDeltaOld > BiasDelta)))
                then begin
                  if FAdaptConstStep > 0 then AdaptBias := AdaptBias - FAdaptConstStep
                                         else AdaptBias := AdaptBias / FAdaptConstFactor;
                  if AdaptBias < 2.0 then AdaptBias := 2.0;
                end
                else begin
                  if FAdaptConstStep > 0 then AdaptBias := AdaptBias + FAdaptConstStep
                                         else AdaptBias := AdaptBias * FAdaptConstFactor;
                  if AdaptBias > FAdaptConst then AdaptBias := FAdaptConst;
                end;
                BiasDelta2Old := BiasDeltaOld;
              end;
            end;
            Bias := Bias + BiasDelta + FMomentum * BiasDeltaOld;
            BiasDeltaOld := BiasDelta;
          end;
          if (lpSlope in FLearnParams) then begin { learning slope? }
            if FAdaptConst > 1 then begin { using adaptive algorithm? }
              SlopeDelta := SlopeDeltaOld + 1.0 / AdaptSlope * (SlopeDelta - SlopeDeltaOld);
              if (FAdaptConstFactor > 1) or (FAdaptConstStep > 0) then begin { using variable adaptation constant? }
                if (((SlopeDelta2Old <= SlopeDeltaOld) and (SlopeDeltaOld <= SlopeDelta)) or
                    ((SlopeDelta2Old > SlopeDeltaOld) and (SlopeDeltaOld > SlopeDelta)))
                then begin
                  if FAdaptConstStep > 0 then AdaptSlope := AdaptSlope - FAdaptConstStep
                                         else AdaptSlope := AdaptSlope / FAdaptConstFactor;
                  if AdaptSlope < 2.0 then AdaptSlope := 2.0;
                end
                else begin
                  if FAdaptConstStep > 0 then AdaptSlope := AdaptSlope + FAdaptConstStep
                                         else AdaptSlope := AdaptSlope * FAdaptConstFactor;
                  if AdaptSlope > FAdaptConst then AdaptSlope := FAdaptConst;
                end;
                SlopeDelta2Old := SlopeDeltaOld;
              end;
            end;
            Slope := Slope + SlopeDelta + FMomentum * SlopeDeltaOld;
            SlopeDeltaOld := SlopeDelta;
          end;
        end;
end;

procedure TBackPropNet.SetInputNo(const Value: Integer);
var I: Integer;
begin
  if Value <> FInputNo then begin
    if Value <= 0 then
      raise Exception.Create('Netz mu mindestens einen Eingang besitzen');
    FInputNo := Value;
    with Items[0] do
      for I := 0 to (NeuronNo - 1) do
        Items[I].InputNo := FInputNo;
  end;
end;

procedure TBackPropNet.SetInputs(InpVector: TRealArray);
var I, J: Integer;
begin
  with Items[0] do
    for I := 0 to (NeuronNo - 1) do
      with Items[I] do begin
        for J := 0 to (InputNo - 1) do
          Inputs[J] := InpVector[J];
        CalcState;
      end;
end;

procedure TBackPropNet.SetInterval(const Value: Integer);
begin
  if Value <> FInterval then
    if Value <= 0 then
      raise Exception.Create('''Interval'' mu ein Integerwert grer Null sein');
    FInterval := Value;
end;

procedure TBackPropNet.SetLayer(const Value: TLayer);
begin
end;

procedure TBackPropNet.SetLayerNo(const Value: Integer);
var I: Integer;
begin
  if Value <> FLayerNo then begin
    if Value <= 0 then
      raise Exception.Create('Netz mu mindestens eine Schicht enthalten');
    while FLayerNo > Value do
      DeleteLayer(FLayerNo - 1);
    while FLayerNo < Value do
      InsertLayer(FLayerNo,1);
    for I := 0 to (FLayerNo - 1) do
      Items[I].Selected := True;
  end;
end;

procedure TBackPropNet.SetLearnParams(const Value: TLearnParams);
var I, J: Integer;
begin
  if Value <> FLearnParams then begin
    FLearnParams := Value;
    for I := 0 to (FLayerNo - 1) do
      with Items[I] do
        for J := 0 to (NeuronNo - 1) do
          Items[J].LearnParams := Value;
  end;
end;

procedure TBackPropNet.SetLearnRate(const Value: Extended);
begin
  if Value <> FLearnRate then begin
    FLearnRate := Value;
    FLearnRateAdapt := Value;
  end;
end;

procedure TBackPropNet.SetLearnRateDec(const Value: Extended);
begin
  if Value <> FLearnRateDec then begin
    if Value > 1 then
      raise Exception.Create('''LearnRateDec'' mu ein Gleitkommawert kleiner oder gleich Eins sein');
    FLearnRateDec := Value;
  end;
end;

procedure TBackPropNet.SetLearnRateInc(const Value: Extended);
begin
  if Value <> FLearnRateInc then begin
    if Value < 1 then
      raise Exception.Create('''LearnRateInc'' mu ein Gleitkommawert grer oder gleich Eins sein');
    FLearnRateInc := Value;
  end;
end;

procedure TBackPropNet.SetMomentum(const Value: Extended);
begin
  if Value <> FMomentum then begin
    if (Value < 0.0) or (Value > 1.0) then
      raise Exception.Create('''Momentum'' mu ein Gleitkommawert im Intervall [0,1] sein');
    FMomentum := Value;
    if FMomentum <> 0 then FAdaptConst := 0;
  end;
end;

procedure TBackPropNet.SetOutputs(OutpVector: TRealArray);
var I: Integer;
begin
  with Items[FLayerNo - 1] do
    for I := 0 to (NeuronNo - 1) do
      OutpVector[I] := Items[I].Output;
end;

procedure TBackPropNet.SetParamsMax(const Value: Extended);
var I, J: Integer;
begin
  if Value <> FParamsMax then begin
    if Value <= FParamsMin then
      raise Exception.Create('''ParamsMax'' mu ein Gleitkommawert und grer als ''ParamsMin'' sein');
    FParamsMax := Value;
    for I := 0 to (FLayerNo - 1) do
      with Items[I] do
        for J := 0 to (NeuronNo - 1) do
          Items[J].ParamsMax := FParamsMax;
  end;
end;

procedure TBackPropNet.SetParamsMin(const Value: Extended);
var I, J: Integer;
begin
  if Value <> FParamsMin then begin
    if Value >= FParamsMax then
      raise Exception.Create('''ParamsMin'' mu ein Gleitkommawert und kleiner als ''ParamsMax'' sein');
    FParamsMin := Value;
    for I := 0 to (FLayerNo - 1) do
      with Items[I] do
        for J := 0 to (NeuronNo - 1) do
          Items[J].ParamsMin := FParamsMin;
  end;
end;

procedure TBackPropNet.SetPatternSet(Value: TSet);
begin
  FPatternSet.Assign(Value);
end;

procedure TBackPropNet.SetRecallSet(Value: TSet);
begin
  FRecallSet.Assign(Value);
end;

procedure TBackPropNet.SetTargets(TargVector: TRealArray);
var I: Integer;
begin
  with Items[FLayerNo - 1] do
    for I := 0 to (NeuronNo - 1) do
      Items[I].Target := TargVector[I];
end;


{ methods of TPatternDialog }
constructor TPatternDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPatternSet := TSet.Create(0, 0, 0);
  FTitle := 'Patternset';
end;

destructor TPatternDialog.Destroy;
begin
  FPatternSet.Free;
  inherited Destroy;
end;

function TPatternDialog.Execute: Boolean;
var Dlg: TPatternEdit;
begin
  Result := False;

  Dlg := TPatternEdit.CreateWithParam(Application, FNeuralNet,
                                      FPatternSet, FSetType, FTitle);
  try
    if Dlg.ShowModal = mrOK then Result := True;
  finally
    Dlg.Free;
  end;
end;

procedure TPatternDialog.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FNeuralNet) and (Operation = opRemove) then
    FNeuralNet := nil;
  if (AComponent is TBackPropNet) and (Operation = opInsert) and (FNeuralNet = nil) then
    FNeuralNet := AComponent as TBackPropNet;
end;

procedure TPatternDialog.SetSetType(Value: TSetType);
begin
  if Value <> FSetType then begin
    FSetType := Value;
    if Value = stPatternSet then FTitle := 'Patternset'
                            else FTitle := 'Recallset';
  end;
end;

procedure TPatternDialog.SetPatternSet(const Value: TSet);
begin
  FPatternSet.Assign(Value);
end;


{ methods of TPatternProperty }
procedure TPatternProperty.Edit;
var Dlg: TPatternDialog; Net: TBackPropNet;
begin
  Net := GetComponent(0) as TBackPropNet;
  Dlg := TPatternDialog.Create(Application);
  with Dlg do
  try
    NeuralNet := Net;
    if GetName = 'PatternSet' then
      PatternSet := Net.PatternSet
    else begin
      PatternSet := Net.RecallSet;
      SetType := stRecallSet;
    end;
    if Execute then begin
      if GetName = 'PatternSet' then Net.PatternSet.Assign(PatternSet)
                                else Net.RecallSet.Assign(PatternSet);
      Designer.Modified;
    end;
  finally
    Dlg.Free;
  end;
end;

function TPatternProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paReadOnly];
end;

function TPatternProperty.GetValue: string;
var I: Integer;
begin
  if GetName = 'PatternSet' then
    I := (GetComponent(0) as TBackPropNet).PatternSet.PatternNo
  else
    I := (GetComponent(0) as TBackPropNet).RecallSet.PatternNo;
  Result := 'PatternNo:  ' + IntToStr(I);
end;


{ methods of TLayerIntProperty }
function TLayerIntProperty.AllEqual: Boolean;
var Layer: TLayer; Val: Integer;
begin
  Result := False;
  Layer := TLayer(GetComponent(0)).Net[0];
  while (not Layer.Selected) do
    Layer := Layer.NextLayer;
  Val := GetOrdProp(Layer, GetPropInfo);
  Layer := Layer.NextLayer;
  while Layer <> nil do begin
    if Layer.Selected and (Val <> GetOrdProp(Layer, GetPropInfo)) then Exit;
    Layer := Layer.NextLayer;
  end;
  Result := True;
end;

function TLayerIntProperty.GetValue: string;
var Layer: TLayer;
begin
  Layer := TLayer(GetComponent(0)).Net[0];
  while (not Layer.Selected) do
    Layer := Layer.NextLayer;
  Result := IntToStr(GetOrdProp(Layer, GetPropInfo));
end;

procedure TLayerIntProperty.SetValue(const Value: string);
var Layer: TLayer;
begin
  Layer := TLayer(GetComponent(0)).Net[0];
  while Layer <> nil do begin
    if Layer.Selected then
      SetOrdProp(Layer, GetPropInfo, StrToInt(Value));
    Layer := Layer.NextLayer;
  end;
  Modified;
end;


{ methods of TLayerEnumProperty }
function TLayerEnumProperty.AllEqual: Boolean;
var Layer: TLayer; Val: Integer;
begin
  Result := False;
  Layer := TLayer(GetComponent(0)).Net[0];
  while (not Layer.Selected) do
    Layer := Layer.NextLayer;
  Val := GetOrdProp(Layer, GetPropInfo);
  Layer := Layer.NextLayer;
  while Layer <> nil do begin
    if Layer.Selected and (Val <> GetOrdProp(Layer, GetPropInfo)) then Exit;
    Layer := Layer.NextLayer;
  end;
  Result := True;
end;

function TLayerEnumProperty.GetValue: string;
var Layer: TLayer;
begin
  Layer := TLayer(GetComponent(0)).Net[0];
  while (not Layer.Selected) do
    Layer := Layer.NextLayer;
  Result := GetEnumName(GetPropType, GetOrdProp(Layer, GetPropInfo));
end;

procedure TLayerEnumProperty.SetValue(const Value: string);
var Layer: TLayer;
begin
  Layer := TLayer(GetComponent(0)).Net[0];
  while Layer <> nil do begin
    if Layer.Selected then
      SetOrdProp(Layer, GetPropInfo, GetEnumValue(GetPropType, Value));
    Layer := Layer.NextLayer;
  end;
  Modified;
end;


{ methods of TLayerFloatProperty }
function TLayerFloatProperty.AllEqual: Boolean;
var Layer: TLayer; Val: Extended;
begin
  Result := False;
  Layer := TLayer(GetComponent(0)).Net[0];
  while (not Layer.Selected) do
    Layer := Layer.NextLayer;
  Val := GetFloatProp(Layer, GetPropInfo);
  Layer := Layer.NextLayer;
  while Layer <> nil do begin
    if Layer.Selected and (Val <> GetFloatProp(Layer, GetPropInfo)) then Exit;
    Layer := Layer.NextLayer;
  end;
  Result := True;
end;

function TLayerFloatProperty.GetValue: string;
var Layer: TLayer;
const
  Precisions: array[TFloatType] of Integer = (7, 15, 18, 18, 18);
begin
  Layer := TLayer(GetComponent(0)).Net[0];
  while (not Layer.Selected) do
    Layer := Layer.NextLayer;
  Result := FloatToStrF(GetFloatProp(Layer, GetPropInfo), ffGeneral,
    Precisions[GetTypeData(GetPropType)^.FloatType], 0);
end;

procedure TLayerFloatProperty.SetValue(const Value: string);
var Layer: TLayer;
begin
  Layer := TLayer(GetComponent(0)).Net[0];
  while Layer <> nil do begin
    if Layer.Selected then
      SetFloatProp(Layer, GetPropInfo, StrToFloat(Value));
    Layer := Layer.NextLayer;
  end;
  Modified;
end;


{ methods of TLayerProperty }
procedure TLayerProperty.Edit;
var Dlg: TSelectDialog;
begin
  Dlg := TSelectDialog.CreateWithParam(Application, GetComponent(0) as TBackPropNet);
  try
    if Dlg.ShowModal = mrOK then Designer.Modified;
  finally
    Dlg.Free;
  end;
end;

function TLayerProperty.GetValue: string;
var I: Integer; Net: TBackPropNet;
begin
  Result := 'Selected: ';
  Net := GetComponent(0) as TBackPropNet;
  for I := 0 to (Net.LayerNo - 1) do
    if Net[I].Selected then
      if Result = 'Selected: ' then Result := Result + IntToStr(I)
                               else Result := Result + ', ' + IntToStr(I);
end;

function TLayerProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paMultiSelect, paSubProperties, paReadOnly];
end;


{ methods of TLayerComponentEditor }
procedure TLayerComponentEditor.Edit;
begin
  ExecuteVerb(0);
end;

procedure TLayerComponentEditor.ExecuteVerb(Index: Integer);
var
  FileDlg: TNetFileDialog;
  InsertDlg: TInsertDialog;
  Net: TBackPropNet;
begin
  Net := Component as TBackPropNet;
  case Index of
    0..3: begin
            if (Net.LayerNo = 1) and (Index = 2) then begin
              MessageDlg('Kann das Neuron in dieser Schicht nicht lschen.' + #13#10 +
              ' Jede Schicht mu mindestens ein Neuron enthalten', mtError, [mbOK], 0);
              Exit;
            end;
            InsertDlg := TInsertDialog.CreateWithParam(Application, Net, Index);
            try
              if InsertDlg.ShowModal = mrOK then Designer.Modified;
            finally
              InsertDlg.Free;
            end;
          end;
    4..5: begin
            FileDlg := TNetFileDialog.Create(Application);
            with FileDlg do try
              NeuralNet := Net;
              if Index = 5 then DialogMode := dmSave;
              if Execute then try
                if Index = 4 then Net.LoadFromFile(FileName)
                             else Net.SaveToFile(FileName);
              except
                on E: Exception do MessageDlg(E.Message, mtError, [mbOK], 0);
              end;
            finally
              Designer.Modified;
              FileDlg.Free;
            end;
          end;
  end;
end;

function TLayerComponentEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := 'Schicht ein&fgen...';
    1: Result := 'Schicht ls&chen...';
    2: Result := 'Neuron e&infgen...';
    3: Result := 'Neuron &lschen...';
    4: Result := 'La&den...';
    5: Result := 'S&peichern...';
  end;
end;

function TLayerComponentEditor.GetVerbCount: Integer;
begin
  Result := 6;
end;


{ registration of components and editors }
procedure Register;
begin
  RegisterComponents('NeuroNet', [TBackPropNet, TPatternDialog]);
  RegisterPropertyEditor(TypeInfo(TSet), TBackPropNet, '', TPatternProperty);
  RegisterPropertyEditor(TypeInfo(Integer), TLayer, '', TLayerIntProperty);
  RegisterPropertyEditor(TypeInfo(TFunctTypeAct), TLayer, 'ActFunction', TLayerEnumProperty);
  RegisterPropertyEditor(TypeInfo(TFunctTypeOutp), TLayer, 'OutpFunction', TLayerEnumProperty);
  RegisterPropertyEditor(TypeInfo(TFunctTypeProp), TLayer, 'PropFunction', TLayerEnumProperty);
  RegisterPropertyEditor(TypeInfo(Extended), TLayer, '', TLayerFloatProperty);
  RegisterPropertyEditor(TypeInfo(TLayer), TBackPropNet, 'Layer', TLayerProperty);
  RegisterComponentEditor(TBackPropNet, TLayerComponentEditor);
end;

end.

