unit FileDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Forms, Dialogs, Dlgs, FileCtrl,
  Ole2, ShellApi, ShlObj, BackPropNet, CommCtrl, CommDlg, FiltEdit, DsgnIntf;

{ TFileDialog }
type
  TDialogMode = (dmOpen, dmSave);
  TFileEditStyle = (fsEdit, fsComboBox);

  TOpenOption = (ofReadOnly, ofOverwritePrompt, ofHideReadOnly, ofNoChangeDir, ofShowHelp,
                 ofNoValidate, ofAllowMultiSelect, ofExtensionDifferent, ofPathMustExist,
                 ofFileMustExist, ofCreatePrompt, ofShareAware, ofNoReadOnlyReturn, ofNoTestFileCreate,
                 ofNoNetworkButton, ofNoLongNames, ofOldStyleDialog, ofNoDereferenceLinks);
  TOpenOptions = set of TOpenOption;
  TCommandNotifyEvent = procedure(Sender: TObject; hCtrl: HWND; idCtrl: Integer) of object;

  TFileDialog = class(TCommonDialog)
  private
    { property data }
    FCenter: Boolean;                  { center dialog }
    FDefaultExt: string;
    FDialogMode: TDialogMode;          { open dialog or save dialog? }
    FDirectory: string;
    FFileEditStyle: TFileEditStyle;
    FFileName: TFileName;
    FFiles: TStrings;
    FFilter: string;
    FFilterIndex: Integer;
    FHistoryList: TStrings;
    FHCommonDlg: HWND;                 { handle of custom/ common dialog }
    FHCustomDlg: HWND;
    FOptions: TOpenOptions;
    FShowTemplate: Boolean;            { show custom dialog? }
    FTemplateName: string;             { dialog template name }
    FTitle: string;
    { property methods }
    function GetCtrlEnabled(Id: Integer): Boolean;
    function GetCtrlText(Id: Integer): string;
    function GetCtrlVisible(Id: Integer): Boolean;
    function GetDirectory: string;     { returns path and name of the currently opened folder }
    function GetFile: string;          { returns selected file (not including the path) }
    function GetFileName: TFileName;   { returns path and filename of the selected file }
    procedure SetCtrlEnabled(Id: Integer; Value: Boolean);
    procedure SetCtrlText(Id: Integer; const Value: string);
    procedure SetCtrlVisible(Id: Integer; Value: Boolean);
    procedure SetDefaultExt(const Value: string);
    procedure SetDirectory(const Value: string);
    procedure SetHistoryList(Value: TStrings);
    procedure SetShowTemplate(Value: Boolean);
    procedure SetTitle(const Value: string);
  protected
    CustomSize: TPoint;                { store size of custom/ common dialog }
    OrigSize: TPoint;
    Executed: Boolean;                 { is true during run time }
    { event handlers }
    FOnCommand: TCommandNotifyEvent;   { dialog controls are used }
    FOnInit: TNotifyEvent;             { system starts to arrange common controls }
    FOnInitDone: TNotifyEvent;         { system has finished arranging common controls }
    FOnFileChange: TNotifyEvent;       { new file is selected in file list }
    FOnFileOk: TNotifyEvent;           { dialog is about to close after OK- button is pressed }
    FOnFolderChange: TNotifyEvent;     { new folder is selected }
    FOnHelp: TNotifyEvent;             { help button is pressed }
    FOnTypeChange: TNotifyEvent;       { new file typ is selected in filter box}
    function DoExecute(Func: Pointer): Bool;
    procedure Command(hCtrl: HWND; idCtrl: Integer); virtual;
    procedure FileChange; virtual;
    procedure FileOk; virtual;
    procedure FolderChange; virtual;
    procedure Help; virtual;
    procedure Init; virtual;
    procedure InitDone; virtual;
    procedure TypeChange; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: Boolean; virtual;
    property AFile: string read GetFile;
    property CtrlEnabled[Id: Integer]: Boolean read GetCtrlEnabled write SetCtrlEnabled;
    property CtrlText[Id: Integer]: string read GetCtrlText write SetCtrlText;
    property CtrlVisible[Id: Integer]: Boolean read GetCtrlVisible write SetCtrlVisible;
    property Files: TStrings read FFiles;
    property HCommonDlg: HWND read FHCommonDlg;
    property HCustomDlg: HWND read FHCustomDlg;
  published
    property Center: Boolean read FCenter write FCenter default True;
    property DefaultExt: string read FDefaultExt write SetDefaultExt;
    property DialogMode: TDialogMode read FDialogMode write FDialogMode default dmOpen;
    property Directory: string read GetDirectory write SetDirectory;
    property FileEditStyle: TFileEditStyle read FFileEditStyle write FFileEditStyle
                            default fsEdit;
    property FileName: TFileName read GetFileName write FFileName;
    property Filter: string read FFilter write FFilter;
    property FilterIndex: Integer read FFilterIndex write FFilterIndex default 1;
    property HistoryList: TStrings read FHistoryList write SetHistoryList;
    property OnCommand: TCommandNotifyEvent read FOnCommand write FOnCommand;
    property OnFileChange: TNotifyEvent read FOnFileChange write FOnFileChange;
    property OnFileOk: TNotifyEvent read FOnFileOk write FOnFileOk;
    property OnFolderChange: TNotifyEvent read FOnFolderChange write FOnFolderChange;
    property OnHelp: TNotifyEvent read FOnHelp write FOnHelp;
    property OnInitDone: TNotifyEvent read FOnInitDone write FOnInitDone;
    property OnTypeChange: TNotifyEvent read FOnTypeChange write FOnTypeChange;
    property Options: TOpenOptions read FOptions write FOptions
                      default [ofOverwritePrompt, ofHideReadOnly, ofShowHelp];
    property ShowTemplate: Boolean read FShowTemplate write SetShowTemplate default True;
    property TemplateName: string read FTemplateName write FTemplateName;
    property Title: string read FTitle write SetTitle;
  end;


{ TNetFileDialog }
const
  DefaultFilter: string =
    'Netzdateien (*.cfg; *.lrn; *.net; *.ptn; *.rcl)|*.cfg;*.lrn;*.net;*.ptn;*.rcl|' +
    'Konfiguration (*.cfg)|*.cfg|' +
    'Gelernte Parameter (*.lrn)|*.lrn|' +
    'Gesamtes Netz (*.net)|*.net|' +
    'Patternset (*.ptn)|*.ptn|' +
    'Recallset (*.rcl)|*.rcl|' +
    'Textdateien (*.csv, *.prn; *.txt)|*.csv;*.prn;*.txt|' +
    'Alle Dateien (*.*)|*.*|';
  DefaultTemplate: string = 'NetDialog';

type
  TNetFileDialog = class(TFileDialog)
  private
    DlgTag: Integer;                 { used to distinguish several dialogs }
    DeleteSource: Boolean;           { stores state of chx10 }
    Target: string;                  { stores text of edt10 }
    TypeIndex: Integer;              { stores index of cmb10 }
    { property data }
    FNeuralNet: TBackPropNet;        { points to neural net }
    { property methods }
    function IsFilterStored: Boolean;
    function IsTemplateNameStored: Boolean;
  protected
    procedure AddFiles(Net: TBackPropNet);
    procedure Command(hCtrl: HWND; idCtrl: Integer); override;
    procedure ConvertFiles;
    procedure EditBinary(const FileName: string); virtual;
    procedure FileChange; override;
    procedure FileOk; override;
    procedure Init; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure ProcessFile;
    procedure ProcessFiles;
  public
    constructor Create(AOwner: TComponent); override;
    procedure ConvertAscFileToBinFile(const AscFileName, BinFileName: string); virtual;
    function Execute: Boolean; override;
  published
    property Filter stored IsFilterStored;
    property NeuralNet: TBackPropNet read FNeuralNet write FNeuralNet;
    property TemplateName stored IsTemplateNameStored;
  end;

function BrowseFolder(Wnd: HWND; Root: Integer; Flags: UINT; const Title,
                      Path: string): string;
procedure ConvertAscFileToStream(const AscFileName: string; Stream: TStream);

procedure Register;

implementation
{$R NETDLG.RES}

uses Edit;

{ hook function for SHBrowseForFolder }
function BrowseFolderHook(Wnd: HWND; Msg: UINT; LParam, lpData: LPARAM): UINT; stdcall;
begin
  Result := 0;
  if Msg = BFFM_INITIALIZED then begin
   SendMessage(Wnd, BFFM_SETSELECTION, 0, lpData);
  end;
end;

{ creates browse dialog, returnes selected folder }
function BrowseFolder(Wnd: HWND; Root: Integer; Flags: UINT; const Title,
                      Path: string): string;
var
  Allocator: IMalloc;
  BrowseInfo: TBrowseInfo;
  Desktop: IShellFolder;
  IdCurrent, IdRoot: PItemIdList;
  NotUsed: ULONG;
  UniStr: array[0..MAX_PATH] of WideChar;
begin
  SetLength(Result, MAX_PATH);
  { get shell allocator }
  if Succeeded(SHGetMalloc(Allocator)) then
  try
    { create IShellFolder object for virtual desktop }
    if Succeeded(SHGetDesktopFolder(Desktop)) then
    try
      StringToWideChar(Path, UniStr, SizeOf(UniStr));
      { get item identifer of root to browse from }
      SHGetSpecialFolderLocation(Wnd, Root, IdRoot);
      with BrowseInfo do begin
        hWndOwner := Wnd;
        pidlRoot := IdRoot;
        pszDisplayName := PChar(Result);
        lpszTitle := 'Ausgabeverzeichnis fr Konvertierung whlen:';
        ulFlags := Flags;
        { convert current path to item identifer }
        if Succeeded(Desktop.ParseDisplayName(Wnd, nil, UniStr, NotUsed,
                     IdCurrent, NotUsed)) then begin
          lpfn := BrowseFolderHook;
          lParam := LongInt(IdCurrent);
        end
        else lpfn := nil;
      end;
      { execute the dialog }
      IdRoot := SHBrowseForFolder(@BrowseInfo);
      { convert to real path }
      if Assigned(IdRoot) then SHGetPathFromIDList(IdRoot, PChar(Result))
                          else Result := Path;
    finally
      Allocator.Free(IdCurrent);
      Allocator.Free(IdRoot);
      Desktop.Release;
    end;
  finally
    Allocator.Release;
  end;
end;

{ centers dialog on screen }
procedure CenterWindow(Wnd: HWND);
var Rect: TRect;
begin
  GetWindowRect(Wnd, Rect);
  SetWindowPos(Wnd, 0, (GetSystemMetrics(SM_CXSCREEN) - Rect.Right + Rect.Left) div 2,
               (GetSystemMetrics(SM_CYSCREEN) - Rect.Bottom + Rect.Top) div 2,
               0, 0, SWP_NOACTIVATE or SWP_NOSIZE or SWP_NOZORDER);
end;

{ estimates and stores size of common and custom dialog }
procedure SizeWindow(Obj: TFileDialog);
var OrigRect, CustRect: TRect;
begin
  { get size of custom dialog }
  GetWindowRect(Obj.HCommonDlg, CustRect);
  { get size of common dialog (placeholder stc32) if exists }
  if GetDlgItem(Obj.HCustomDlg, stc32) <> 0 then
    GetWindowRect(GetDlgItem(Obj.HCustomDlg, stc32), OrigRect)
  else CopyRect(OrigRect, CustRect);
  Obj.CustomSize := Point(CustRect.Right - CustRect.Left, CustRect.Bottom - CustRect.Top);
  Obj.OrigSize := Point(OrigRect.Right - CustRect.Left, OrigRect.Bottom - CustRect.Top);
  Obj.SetShowTemplate(Obj.FShowTemplate);
end;

{ hook function for custom dialog }
function ExplorerHook(Wnd: HWND; Msg: UINT; WParam: WPARAM; LParam: LPARAM): UINT; stdcall;
var
  Ofn : ^TOpenFileName;
  Obj : TFileDialog;
begin
  Result := 0;  { allow standard processing }
  if (Msg = WM_INITDIALOG) then begin
    Ofn := Pointer(LParam);
    SetProp(Wnd, 'TCOD', Ofn.lCustData);
    Obj := Pointer(Ofn.lCustData);
    Obj.FHCustomDlg := Wnd;
    Obj.FHCommonDlg := GetParent(Wnd);
    Obj.Init;
  end;
  Obj := Pointer(GetProp(Wnd, 'TCOD'));  { get pointer to object }
  if (Msg = WM_NCDESTROY) then begin
    RemoveProp(Wnd, 'TCOD');  { clean up }
  end;
  if (Msg = WM_COMMAND) then Obj.Command(LParam, LOWORD(WParam));
  if (Msg = WM_NOTIFY) then begin { distribute notifications }
    case POFNotify(LParam)^.hdr.code of  { dispatch each event }
      CDN_INITDONE    : begin
                          SizeWindow(Obj);  { store and set dialog size }
                          if Obj.FCenter then  { center if requested }
                            CenterWindow(GetWindowLong(Wnd, GWL_HWNDPARENT));
                          Obj.InitDone;
                         end;
      CDN_FILEOK      : Obj.FileOk;
      CDN_SELCHANGE   : Obj.FileChange;
      CDN_FOLDERCHANGE: Obj.FolderChange;
      CDN_TYPECHANGE  : Obj.TypeChange;
      CDN_HELP        : Obj.Help;
    end;
  end;
end;


{ methods of TFileDialog }
constructor TFileDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCenter := True;
  FFiles := TStringList.Create;
  FFilterIndex := 1;
  FFileEditStyle := fsEdit;
  FHistoryList := TStringList.Create;
  FOptions := [ofOverwritePrompt, ofHideReadOnly, ofShowHelp];
  FShowTemplate := True;
end;

destructor TFileDialog.Destroy;
begin
  FFiles.Free;
  FHistoryList.Free;
  inherited Destroy;
end;

procedure TFileDialog.Command(hCtrl: HWND; idCtrl: Integer);
begin
  if Assigned(FOnCommand) then
    FOnCommand(Self, hCtrl, idCtrl);
end;

{ more or less obtained from Dialogs.pas }
function TFileDialog.DoExecute(Func: Pointer): BOOL;
const
  MultiSelectBufferSize = 8192;
  OpenOptions: array [TOpenOption] of Longint = (
    OFN_READONLY, OFN_OVERWRITEPROMPT, OFN_HIDEREADONLY, OFN_NOCHANGEDIR,
    OFN_SHOWHELP, OFN_NOVALIDATE, OFN_ALLOWMULTISELECT, OFN_EXTENSIONDIFFERENT,
    OFN_PATHMUSTEXIST, OFN_FILEMUSTEXIST, OFN_CREATEPROMPT, OFN_SHAREAWARE,
    OFN_NOREADONLYRETURN, OFN_NOTESTFILECREATE, OFN_NONETWORKBUTTON,
    OFN_NOLONGNAMES, OFN_EXPLORER, OFN_NODEREFERENCELINKS);
var
  CDefaultExt: array[0..3] of Char;
  P: PChar;
  OpenFilename: TOpenFilename;
  Option: TOpenOption;

  function AllocFilterStr(const S: string): PChar;
  var P: PChar;
  begin
    Result := nil;
    if S <> '' then begin
      Result := StrCopy(StrAlloc(Length(S) + 2), PChar(S));
      P := Result;
      while P^ <> #0 do begin
        if P^ = '|' then P^ := #0;
        Inc(P);
      end;
      Inc(P);
      P^ := #0;
    end;
  end;

  function ExtractFileName(P: PChar; var S: string): PChar;
  var Separator: Char;
  begin
    Separator := #0;
    if ofOldStyleDialog in FOptions then Separator := ' ';
    Result := P;
    while (Result[0] <> #0) and (Result[0] <> Separator) do Inc(Result);
    SetString(S, P, Result - P);
    if Result[0] = Separator then Inc(Result);
  end;

  procedure ExtractFileNames(P: PChar);
  var DirName, FileName: string;
  begin
    P := ExtractFileName(P, DirName);
    P := ExtractFileName(P, FileName);
    if FileName = '' then FFiles.Add(DirName)
    else repeat
      if (FileName[1] <> '\') and ((Length(FileName) <= 3) or
         (FileName[2] <> ':') or (FileName[3] <> '\')) then
        FileName := DirName + '\' + FileName;
      FFiles.Add(FileName);
      P := ExtractFileName(P, FileName);
    until FileName = '';
  end;

begin
  FFiles.Clear;
  FillChar(OpenFileName, SizeOf(OpenFileName), 0);
  with OpenFilename do
  try
    lStructSize := SizeOf(TOpenFilename);
    hInstance := System.HInstance;
    lpstrFilter := AllocFilterStr(FFilter);
    nFilterIndex := FFilterIndex;
    if ofAllowMultiSelect in FOptions then
      nMaxFile := MultiSelectBufferSize
    else nMaxFile := MAX_PATH;
    GetMem(lpstrFile, nMaxFile + 2);
    FillChar(lpstrFile^, nMaxFile + 2, 0);
    StrLCopy(lpstrFile, PChar(FFileName), nMaxFile);
    lpstrInitialDir := PChar(FDirectory);
    lpstrTitle := PChar(FTitle);
    Flags := OFN_ENABLEHOOK;  { always enable hook }
    for Option := Low(Option) to High(Option) do
      if Option in FOptions then
        Flags := Flags or OpenOptions[Option];
    Flags := Flags or OFN_EXPLORER; { always use new style }
    if FDefaultExt <> '' then begin
      P := PChar(FDefaultExt);
      lpstrDefExt := StrLCopy(CDefaultExt, P, 3)
    end;
    lpfnHook := ExplorerHook;  { add custom hook }
    if FTemplateName <> '' then begin  { add custom resource }
      if FindResource(System.HInstance, PChar(FTemplateName), RT_DIALOG) = 0 then
        MessageDlg('Kann Template ''' + FTemplateName + ''' nicht laden!', mtError, [mbOK], 0)
      else begin
        lpTemplateName := PChar(FTemplateName);
        Flags := Flags or OFN_ENABLETEMPLATE;
      end;
    end;
    lCustData:=LongInt(Self);  { permit hook function to find file dialog }
    hWndOwner := Application.Handle;
    Result := TaskModalDialog(Func, OpenFileName);
    if Result then begin
      if ofAllowMultiSelect in FOptions then begin
        ExtractFileNames(lpstrFile);
        FFileName := FFiles[0];
      end
      else begin
        ExtractFileName(lpstrFile, FFileName);
        FFiles.Add(FFileName);
      end;
      if (Flags and OFN_EXTENSIONDIFFERENT) <> 0 then
        Include(FOptions, ofExtensionDifferent) else
        Exclude(FOptions, ofExtensionDifferent);
      if (Flags and OFN_READONLY) <> 0 then
        Include(FOptions, ofReadOnly) else
        Exclude(FOptions, ofReadOnly);
      FFilterIndex := nFilterIndex;
    end;
  finally
    if lpstrFile <> nil then FreeMem(lpstrFile, nMaxFile + 2);
    if lpstrFilter <> nil then StrDispose(lpstrFilter);
  end;
end;

function TFileDialog.Execute: Boolean;
begin
  Executed := True;
  try
    if DialogMode = dmOpen then Result := DoExecute(@GetOpenFileName)
                           else Result := DoExecute(@GetSaveFileName);
  finally
    Executed := False;
  end;
end;

procedure TFileDialog.FileChange;
begin
  if Assigned(FONFileChange) then FOnFileChange(Self);
end;

procedure TFileDialog.FileOk;
begin
  if Assigned(FOnFileOk) then FOnFileOk(Self);
end;

procedure TFileDialog.FolderChange;
begin
  if Assigned(FOnFolderChange) then FOnFolderChange(Self);
end;

function TFileDialog.GetCtrlEnabled(Id: Integer): Boolean;
var Wnd: HWND;
begin
  if GetDlgItem(FHCustomDlg, Id) <> 0 then Wnd := FHCustomDlg
                                      else Wnd := FHCommonDlg;
  Result := IsWindowEnabled(GetDlgItem(Wnd, Id));
end;

function TFileDialog.GetCtrlText(Id: Integer): string;
var I: Integer; Wnd: HWND;
begin
  if GetDlgItem(FHCustomDlg, Id) <> 0 then Wnd := FHCustomDlg
                                      else Wnd := FHCommonDlg;
  SetLength(Result, 128);
  I := GetDlgItemText(Wnd, Id, PChar(Result), 128);
  SetLength(Result, I);
end;

function TFileDialog.GetCtrlVisible(Id: Integer): Boolean;
var Wnd: HWND;
begin
  if GetDlgItem(FHCustomDlg, Id) <> 0 then Wnd := FHCustomDlg
                                      else Wnd := FHCommonDlg;
  Result := IsWindowVisible(GetDlgItem(Wnd, Id));
end;

function TFileDialog.GetDirectory: string;
var I: Integer;
begin
  if Executed then begin
    SetLength(Result, 256);
    I := SendMessage(FHCommonDlg, CDM_GetFolderPath, 256, LongInt(PChar(Result)));
    SetLength(Result, I - 1);
  end
  else Result := FDirectory;
end;

function TFileDialog.GetFile: string;
var I: Integer;
begin
   SetLength(Result, 256);
   I := SendMessage(FHCommonDlg, CDM_GETSPEC, 256, LongInt(PChar(Result)));
   SetLength(Result, I - 1);
end;

function TFileDialog.GetFileName: TFileName;
var I: Integer;
begin
   if Executed then begin
     SetLength(Result, 512);
     I := SendMessage(FHCommonDlg, CDM_GETFILEPATH, 512, LongInt(PChar(Result)));
     SetLength(Result, I - 1);
   end
   else Result := FFileName;
end;

procedure TFileDialog.Help;
begin
  if Assigned(FOnHelp) then FOnHelp(Self);
end;

procedure TFileDialog.Init;
begin
  if Assigned(FOnInit) then FOnInit(Self);
end;

procedure TFileDialog.InitDone;
begin
  if Assigned(FOnInitDone) then FOnInitDone(Self);
end;

procedure TFileDialog.SetCtrlEnabled(Id: Integer; Value: Boolean);
var Wnd: HWND;
begin
  if GetDlgItem(FHCustomDlg, Id) <> 0 then Wnd := FHCustomDlg
                                      else Wnd := FHCommonDlg;
  EnableWindow(GetDlgItem(Wnd, Id), Value);
end;

procedure TFileDialog.SetCtrlText(Id: Integer; const Value: string);
var Wnd: HWND;
begin
  if GetDlgItem(FHCustomDlg, Id) <> 0 then Wnd := FHCustomDlg
                                      else Wnd := FHCommonDlg;
  SetDlgItemText(Wnd, Id, PChar(Value));
end;

procedure TFileDialog.SetCtrlVisible(Id: Integer; Value: Boolean);
const ShowCmd: array[Boolean] of Integer = (SW_HIDE, SW_SHOW);
var Wnd: HWND;
begin
  if GetDlgItem(FHCustomDlg, Id) <> 0 then Wnd := FHCustomDlg
                                      else Wnd := FHCommonDlg;
  ShowWindow(GetDlgItem(Wnd, Id), ShowCmd[Value]);
end;

procedure TFileDialog.SetDefaultExt(const Value: string);
begin
  if Executed then
    SendMessage(FHCommonDlg, CDM_SETDEFEXT, 0, LongInt(PChar(Value)))
  else FDefaultExt := Value;
end;

procedure TFileDialog.SetDirectory(const Value: string);
var L: Integer;
begin
  L := Length(Value);
  if (L > 1) and (Value[L] = '\') and (Value[L - 1] <> ':') then Dec(L);
  FDirectory := Copy(Value, 1, L);
end;

procedure TFileDialog.SetHistoryList(Value: TStrings);
begin
  FHistoryList.Assign(Value);
end;

procedure TFileDialog.SetShowTemplate(Value: Boolean);
begin
  if Value then begin
    SetWindowPos(FHCommonDlg, 0, 0, 0, CustomSize.X, CustomSize.Y,
                 SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOZORDER);
    ShowWindow(FHCustomDlg, SW_SHOW);
  end
  else begin
    SetWindowPos(FHCommonDlg, 0, 0, 0, OrigSize.X, OrigSize.Y,
                 SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOZORDER);
    ShowWindow(FHCustomDlg, SW_HIDE);
  end;
  FShowTemplate := Value;
end;

procedure TFileDialog.SetTitle(const Value: string);
begin
  if Executed then SetWindowText(FHCommonDlg, PChar(Value))
  else FTitle := Value;
end;

procedure TFileDialog.TypeChange;
begin
  if Assigned(FOnTypeChange) then FOnTypeChange(Self);
end;


{ TNetFileDialog }
procedure ConvertAscFileToStream(const AscFileName: string; Stream: TStream);
var
  I, J, ValCount: Integer;
  DoubleVal: Extended;
  Line: string;
  TestSignature: string[9];
  AscFile: System.Text;
begin
  AssignFile(AscFile, AscFileName);
  Reset(AscFile);
  try
    Read(AscFile, TestSignature);  { read FileSignature }
    TestSignature[0] := #9;
    if AnsiCompareText(FileSignature, TestSignature) <> 0 then
      raise Exception.Create('Datei ''' + AscFileName + ''' ist keine gltige Netzdatei!');
    Read(AscFile, I); { read PatternNo }
    Stream.Write(I, SizeOf(I));
    Read(AscFile, J); { read InputNo }
    Stream.Write(J, SizeOf(J));
    ValCount := I * J;
    Read(AscFile, J); { read OutputNo }
    Stream.Write(J, SizeOf(J));
    ValCount := ValCount + I * J; { file must contain ValCount values }
    J := 0;
    while not EOF(AscFile) do begin
      ReadLn(AscFile, Line);
      while Line <> '' do begin
        while Line[1] in [' ', ';', #9] do begin
          Delete(Line, 1, 1);
          if Line = '' then Break;
        end;
        if Line = '' then Break;
        for I := 1 to Length(Line) do
          if Line[I] in [' ', ';', #9] then Break;
        DoubleVal := StrToFloat(Copy(Line, 1, I - 1));
        Stream.Write(DoubleVal, SizeOf(DoubleVal));
        Delete(Line, 1, I);
        Inc(J);
      end;
    end;
    if J <> ValCount then
      raise Exception.Create('Falsche Anzahl an Werten in Datei ''' + AscFileName + ''' !');
    Stream.Seek(0, soFromBeginning);  
  finally
    CloseFile(AscFile);
  end;
end;

{ methods of TNetFileDialog }
constructor TNetFileDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFilter := DefaultFilter;
  FTemplateName := DefaultTemplate;
end;

procedure TNetFileDialog.AddFiles(Net: TBackPropNet);
var
  I: Integer;
  Path: string;

  function DoAdd(const AFile: string): Boolean;
  var
    ASet: TSet;
    Ext: string;
    Stream: TMemoryStream;
  begin
    Result := True;
    Ext := LowerCase(ExtractFileExt(AFile));
    try
      if (Ext <> '.ptn') and (Ext <> '.rcl') then begin  { add text file }
        Stream := TMemoryStream.Create;
        try
          Stream.SetSize(4096);
          ConvertAscFileToStream(AFile, Stream);
          case TypeIndex of
            0: Net.PatternSet.AddFromStream(Stream);
            1: Net.RecallSet.AddFromStream(Stream);
            2: begin  { add to file spezified by 'Target' }
                 ASet := TSet.Create(0, 0, 0);
                 try
                   ASet.LoadFromStream(Stream);
                   ASet.SaveToFile(Target, fmAdd);
                 finally
                   ASet.Free;
                 end;
               end;
          end;
        finally
          Stream.Free;
        end;
      end  
      else case TypeIndex of
        0: Net.PatternSet.LoadFromFile(AFile, fmAdd);
        1: Net.RecallSet.LoadFromFile(AFile, fmAdd);
        2: begin  { add to file spezified by 'Target' }
             ASet := TSet.Create(0, 0, 0);
             try
               ASet.LoadFromFile(AFile, fmReplace);
               ASet.SaveToFile(Target, fmAdd);
             finally
               ASet.Free;
             end;
           end;
      end;
      if DeleteSource then DeleteFile(AFile);
    except
      on E: Exception do begin
        case MessageDlg(E.Message + #13#10#10 + 'Kann Datei ''' + AFile + ''' nicht addieren.',
                        mtError, [mbIGNORE, mbRETRY, mbABORT], 0) of
          mrABORT : Result := False;
          mrRETRY : Result := DoAdd(AFile);
        end;
      end;
    end;
  end;

begin
  if TypeIndex = 2 then begin
    Path := ExtractFileDir(Target);
    if not DirectoryExists(Path) and
      (MessageDlg('Das Verzeichnis ''' + Path + ''' existiert nicht.' + #13#10 +
       'Mchten Sie es erstellen?', mtConfirmation, [mbYes, mbNo], 0) = mrNo) then Exit
    else try
      ForceDirectories(Path);
    except
      MessageDlg('Kann Verzeichnis nicht erstellen.', mtError, [mbOK], 0);
      Exit;
    end;
  end;
  for I := 0 to (FFiles.Count - 1) do
    if not DoAdd(FFiles[I]) then Exit;
end;

procedure TNetFileDialog.Command(hCtrl: HWND; idCtrl: Integer);
var
  Ext: string;
  AddDlg, ConvertDlg, BrowseDlg: TNetFileDialog;
begin
  try
    case idCtrl of  { dispatch commands of template buttons }
      psh10: begin
               AddDlg := TNetFileDialog.Create(Application);  { add dialog }
               with AddDlg do
               try
                 FFilter := 'Patternset (*.ptn)|*.ptn|' +
                            'Recallset (*.rcl)|*.rcl|' +
                            'Text Dateien (*.csv, *.prn; *.txt)|*.csv;*.prn;*.txt|' +
                            'Alle Dateien (*.*)|*.*|';
                 Include(FOptions, ofAllowMultiSelect);
                 FTemplateName := 'AddDialog';
                 FTitle := 'Hinzufgen';
                 if Execute then begin
                   AddFiles(Self.FNeuralNet);
                   if (TypeIndex = 0) or (TypeIndex = 1) then
                     PostMessage(Self.FHCommonDlg, WM_Command, IDCANCEL, LongInt(False))
                   else Self.ProcessFile;
                 end;
               finally
                 AddDlg.Free;
               end;
             end;
      psh11: begin
               ConvertDlg := TNetFileDialog.Create(Application);  { convert dialog }
               with ConvertDlg do
               try
                 FFilter := 'Text Dateien (*.csv, *.prn; *.txt)|*.csv;*.prn;*.txt|' +
                            'Alle Dateien (*.*)|*.*|';
                 Include(FOptions, ofAllowMultiSelect);
                 FTemplateName := 'ConvertDialog';
                 FTitle := 'Konvertieren';
                 if Execute then begin
                   ConvertFiles;
                   Self.ProcessFile;
                 end;
               finally
                 ConvertDlg.Free;
               end;
             end;
      psh12: begin  { edit dialogs }
               Ext := LowerCase(ExtractFileExt(GetFileName));
               if (Ext = '.ptn') or (Ext = '.rcl') then EditBinary(FileName)
               else with TEditForm.Create(Application) do
                 Open(FileName);
             end;
      psh13: if DlgTag = 2 then begin { file browser }
               BrowseDlg := TNetFileDialog.Create(Application);
               with BrowseDlg do
               try
                 FFilter := 'Patternset (*.ptn)|*.ptn|' +
                            'Recallset (*.rcl)|*.rcl|' +
                            'Alle Dateien (*.*)|*.*|';
                 FTemplateName := '';
                 FTitle := 'Ausgabedatei whlen';
                 if Execute then
                   Self.CtrlText[edt13] := FFileName;
               finally
                 BrowseDlg.Free;
               end;
             end
             else CtrlText[edt13] := BrowseFolder(FHCommonDlg, CSIDL_DESKTOP,
                             BIF_RETURNONLYFSDIRS, 'SelectFolder', GetDirectory);
      chx10: DeleteSource := (IsDlgButtonChecked(FHCustomDlg, chx10) = 1);
      cmb10: begin
               TypeIndex := SendDlgItemMessage(FHCustomDlg, cmb10, CB_GETCURSEL, 0, 0);
               if DlgTag = 2 then begin
                 CtrlEnabled[stc13] := (TypeIndex = 2);
                 CtrlEnabled[edt13] := (TypeIndex = 2);
                 CtrlEnabled[psh13] := (TypeIndex = 2);
               end;
             end;
    end;
  except
    on E: Exception do MessageDlg(E.Message, mtError, [mbOK], 0);
  end;
  if Assigned(FOnCommand) then FOnCommand(Self, hCtrl, idCtrl);
end;

procedure TNetFileDialog.ConvertAscFileToBinFile(const AscFileName, BinFileName: string);
var Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    Stream.SetSize(4096);
    Stream.Write(FileSignature[1], 9);
    ConvertAscFileToStream(AscFileName, Stream);
    Stream.SaveToFile(BinFileName);
  finally
    Stream.Free;
  end;
end;

procedure TNetFileDialog.ConvertFiles;
var I: Integer;

  function DoConvert(const aFile: string): Boolean;
  var BinFile, Ext: string;
  begin
    Result := True;
    Ext := ExtractFileExt(aFile);
    BinFile := ExtractFileName(aFile);
    if Ext <> '' then BinFile := Target + Copy(BinFile, 1, Pos(Ext, BinFile) - 1)
                 else BinFile := Target + Binfile;
    if TypeIndex = 0 then
      BinFile := Binfile + '.ptn'
    else BinFile := Binfile + '.rcl';
    try
      ConvertAscFileToBinFile(aFile, BinFile);
      if DeleteSource then DeleteFile(aFile);
    except
      on E: Exception do
        case MessageDlg(E.Message + #13#10#10 + 'Kann Datei ''' + aFile + ''' nicht konvertieren.',
                        mtError, [mbIGNORE, mbRETRY, mbABORT], 0) of
          mrABORT : Result := False;
          mrRETRY : Result := DoConvert(aFile);
        end;
    end;
  end;

begin
  if not DirectoryExists(Target) and
    (MessageDlg('Das Verzeichnis ''' + Target + ''' existiert nicht.' + #13#10 +
     'Mchten Sie es erstellen?', mtConfirmation, [mbYes, mbNo], 0) = mrNo) then Exit
  else try
    ForceDirectories(Target);
  except
    MessageDlg('Kann Verzeichnis nicht erstellen.', mtError, [mbOK], 0);
    Exit;
  end;
  if Target[Length(Target)] <> '\' then Target := Target + '\';
  for I := 0 to (FFiles.Count - 1) do
    if not DoConvert(FFiles[I]) then Exit;
end;

procedure TNetFileDialog.EditBinary(const FileName: string);
var Dlg: TPatternDialog;
begin
  Dlg := TPatternDialog.Create(Application);
  with Dlg do
  try
    NeuralNet := FNeuralNet;
    if LowerCase(ExtractFileExt(FileName)) = '.rcl' then
      SetType := stRecallSet;
    PatternSet.LoadFromFile(FileName, fmReplace);
    if Execute then begin
      PatternSet.SaveToFile(FileName, fmReplace);
      ProcessFile;
    end;
  finally
    Dlg.Free;
  end;
end;

function TNetFileDialog.Execute: Boolean;
begin
  if AnsiCompareText(DefaultTemplate, FTemplateName) = 0 then DlgTag := 1
    else if AnsiCompareText('AddDialog', FTemplateName) = 0 then DlgTag := 2
      else if AnsiCompareText('ConvertDialog', FTemplateName) = 0 then DlgTag := 3
        else DlgTag := 0;
  Result := inherited Execute;
end;

procedure TNetFileDialog.FileChange;
begin
  try
    case DlgTag of
      1   : ProcessFile;
      2..3: ProcessFiles;
    end;  
  except
    on E: Exception do MessageDlg(E.Message, mtError, [mbOK], 0);
  end;
  if Assigned(FONFileChange) then FOnFileChange(Self);
end;

procedure TNetFileDialog.FileOk;
begin
  case DlgTag of
    2..3: Target := CtrlText[edt13];
  end;  
  if Assigned(FOnFileOk) then FOnFileOk(Self);
end;

procedure TNetFileDialog.Init;
begin
  case DlgTag of
    1: if not Assigned(FNeuralNet) then CtrlVisible[edt9] := True;
    2: begin
         CtrlText[ID_OK] := '&Start';
         CtrlText[edt13] := GetCurrentDir;
         SendDlgItemMessage(FHCustomDlg, cmb10, CB_ADDSTRING, 0, LPARAM(PChar('Netz (Patternset)')));
         SendDlgItemMessage(FHCustomDlg, cmb10, CB_ADDSTRING, 0, LPARAM(PChar('Netz (Recallset)')));
         SendDlgItemMessage(FHCustomDlg, cmb10, CB_ADDSTRING, 0, LPARAM(PChar('Datei')));
         SendDlgItemMessage(FHCustomDlg, cmb10, CB_SETCURSEL, 0, 0);
       end;
    3: begin
         CtrlText[ID_OK] := '&Start';
         CtrlText[edt13] := GetCurrentDir;
         SendDlgItemMessage(FHCustomDlg, cmb10, CB_ADDSTRING, 0, LPARAM(PChar('Patternset (*.ptn)')));
         SendDlgItemMessage(FHCustomDlg, cmb10, CB_ADDSTRING, 0, LPARAM(PChar('Recallset (*.rcl)')));
         SendDlgItemMessage(FHCustomDlg, cmb10, CB_SETCURSEL, 0, 0);
       end;
  end;
  if Assigned(FOnInit) then FOnInit(Self);
end;

function TNetFileDialog.IsFilterStored: Boolean;
begin
  Result := AnsiCompareText(DefaultFilter, FFilter) <> 0;
end;

function TNetFileDialog.IsTemplateNameStored: Boolean;
begin
  Result := AnsiCompareText(DefaultTemplate, FTemplateName) <> 0;
end;

procedure TNetFileDialog.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FNeuralNet) and (Operation = opRemove) then
    FNeuralNet := nil;
  if (AComponent is TBackPropNet) and (Operation = opInsert) and (FNeuralNet = nil) then
    FNeuralNet := AComponent as TBackPropNet;
end;

procedure TNetFileDialog.ProcessFiles;
var
  I: Integer;
  FileName, FileNames: string;

  procedure ReadFile(const AFile: string);
  var
    I, J, K: Integer;
    AscFile: System.Text;
    Stream: TFileStream;
    Ext: string;
    TestSignature: string[9];
  begin
    if not FileExists(AFile) then Exit;
    Ext := LowerCase(ExtractFileExt(AFile));
    if (Ext = '.ptn') or (Ext = '.rcl') then begin
      Stream := TFileStream.Create(AFile, fmOpenRead);
      try
        Stream.Read(TestSignature[1], 9);
        TestSignature[0] := #9;
        if AnsiCompareText(TestSignature, FileSignature) <> 0 then Exit;
        Stream.Read(I, SizeOf(I));
        Stream.Read(J, SizeOf(J));
        Stream.Read(K, SizeOf(K));
        Stream.Free;
      except
        Stream.Free;
        Exit;
      end;
    end
    else begin
      AssignFile(AscFile, AFile);
      try
        Reset(AscFile);
        ReadLn(AscFile, TestSignature);
        if AnsiCompareText(TestSignature, FileSignature) <> 0 then Exit;
        Read(AscFile, I);
        Read(AscFile, J);
        Read(AscFile, K);
        CloseFile(AscFile);
      except
        CloseFile(AscFile);
        Exit;
      end;
    end;
    CtrlText[edt10] := IntToStr(StrToInt(CtrlText[edt10]) + I);
    CtrlText[edt11] := IntToStr(StrToInt(CtrlText[edt11]) + J);
    CtrlText[edt12] := IntToStr(StrToInt(CtrlText[edt12]) + K);
  end;

begin
  FileNames := ExtractFileName(GetFileName);
  if Length(FileNames) = 0 then Exit;
  if FileNames[1] = '"' then begin
    CtrlText[grp4] := '';
    CtrlText[edt10] := '0';
    CtrlText[edt11] := '0';
    CtrlText[edt12] := '0';
    while FileNames <> '' do begin
      while FileNames[1] in [' ', '"'] do begin
        Delete(FileNames, 1, 1);
        if FileNames = '' then Break;
      end;
      if FileNames = '' then Break;
      for I := 1 to Length(FileNames) do
        if FileNames[I] in [' ', '"'] then Break;
      FileName := Copy(FileNames, 1, I - 1);
      ReadFile(FileName);
      Delete(FileNames, 1, I);
    end;
  end
  else ProcessFile;
end;


procedure TNetFileDialog.ProcessFile;
var
  AscFile: System.Text;
  I, IntVal, LayNo: Integer;
  NewFile: string;
  Ext, S: string;
  TestSignature: string[9];
  Stream: TStream;

  function ReadInt: string;
  var I: Integer;
  begin
    Read(AscFile, I);
    Result := IntToStr(I);
  end;

begin
  NewFile := GetFileName;
  Ext := LowerCase(ExtractFileExt(NewFile));
  if not FileExists(NewFile) or (Pos('*', NewFile) <> 0) then begin
    CtrlText[grp4] := '';
    CtrlText[edt16] := 'Keine Netzdatei !';
    CtrlVisible[edt16] := True;
    CtrlEnabled[psh12] := False;
    Exit;
  end;
  if (Ext = '.cfg') or (Ext = '.lrn') or (Ext = '.net') or
     (Ext = '.ptn') or (Ext = '.rcl') then begin
    try
      Stream := TFileStream.Create(NewFile, fmOpenRead);
    except
      CtrlText[edt16] := 'Kann Datei ''' + GetFile + ''' nicht ffnen !';
      CtrlVisible[edt16] := True;
      CtrlEnabled[IDOK] := False;
      CtrlEnabled[psh12] := False;
      Exit;
    end;
    try
      Stream.Read(TestSignature[1], 9);
      TestSignature[0] := #9;
      if AnsiCompareText(TestSignature, FileSignature) <> 0 then begin
        CtrlText[edt16] := 'Keine Netzdatei !';
        CtrlVisible[edt16] := True;
        if (FDialogMode = dmOpen) then
          CtrlEnabled[IDOK] := False;
        CtrlEnabled[psh12] := True;
        Exit;
      end;
      CtrlVisible[edt16] := False;
      CtrlEnabled[ID_OK] := True;
      if (Ext = '.ptn') or (Ext = '.rcl') then begin
        CtrlEnabled[psh12] := True;
        CtrlText[stc10] := 'Muster';
        CtrlText[stc12] := 'Ausgnge';
        Stream.Read(IntVal, SizeOf(IntVal));
        CtrlText[edt10] := IntToStr(IntVal);
        Stream.Read(IntVal, SizeOf(IntVal));
        CtrlText[edt11] := IntToStr(IntVal);
        Stream.Read(IntVal, SizeOf(IntVal));
        CtrlText[edt12] := IntToStr(IntVal);
        if Ext = '.ptn' then begin
          CtrlText[grp4] := 'Patternset';
          if Assigned(FNeuralNet) then begin
            CtrlText[edt13] := IntToStr(FNeuralNet.PatternSet.PatternNo);
            CtrlText[edt14] := IntToStr(FNeuralNet.PatternSet.InputNo);
            CtrlText[edt15] := IntToStr(FNeuralNet.PatternSet.OutputNo);
          end;
        end
        else begin
          CtrlText[grp4] := 'Recallset';
          if Assigned(FNeuralNet) then begin
            CtrlText[edt13] := IntToStr(FNeuralNet.RecallSet.PatternNo);
            CtrlText[edt14] := IntToStr(FNeuralNet.RecallSet.InputNo);
            CtrlText[edt15] := IntToStr(FNeuralNet.RecallSet.OutputNo);
          end;
        end;
      end;
      if (Ext = '.lrn') or (Ext = '.cfg') or (Ext = '.net') then begin
        CtrlEnabled[psh12] := False;
        CtrlText[stc10] := 'Schichten';
        CtrlText[stc12] := 'Neuronen';
        Stream.Read(LayNo, SizeOf(LayNo));
        CtrlText[edt10] := IntToStr(LayNo);
        Stream.Read(IntVal, SizeOf(IntVal));
        CtrlText[edt11] := IntToStr(IntVal);
        S := '';
        for I := 0 to (LayNo - 1) do begin
          Stream.Read(IntVal, SizeOf(IntVal));
          S := S + IntToStr(IntVal) + ',';
          CtrlText[edt12] := Copy(S, 1, Length(S) - 1);
        end;
        if Assigned(FNeuralNet) then begin
          CtrlText[edt13] := IntToStr(FNeuralNet.LayerNo);
          CtrlText[edt14] := IntToStr(FNeuralNet.InputNo);
          S := '';
          for I := 0 to (FNeuralNet.LayerNo - 1) do
            S := S + IntToStr(FNeuralNet[I].NeuronNo) + ',';
          CtrlText[edt15] := Copy(S, 1, Length(S) - 1);
        end;
        if Ext = '.cfg' then CtrlText[grp4] := 'Netzkonfiguration'
          else if Ext = '.net' then CtrlText[grp4] := 'Gesamtes Netz'
            else if (Ext = '.lrn') then begin
              CtrlText[grp4] := 'Gelernte Parameter';
              if ((CtrlText[edt10] <> CtrlText[edt13]) or
                  (CtrlText[edt11] <> CtrlText[edt14]) or
                  (CtrlText[edt12] <> CtrlText[edt15])) and
                  (FDialogMode = dmOpen) then
                CtrlEnabled[ID_OK] := False;
            end;
      end;
    finally
      Stream.Free;
    end;
  end
  else begin
    AssignFile(AscFile, NewFile);
    try
      Reset(AscFile);
    except
      CtrlText[edt16] := 'Kann Datei ''' + GetFile + ''' nicht ffnen !';
      CtrlVisible[edt16] := True;
      CtrlEnabled[IDOK] := False;
      CtrlEnabled[psh12] := False;
      Exit;
    end;
    try
      CtrlEnabled[psh12] := True;
      ReadLn(AscFile, TestSignature);
      if AnsiCompareText(TestSignature, FileSignature) <> 0 then begin
        CtrlText[edt16] := 'Keine Netzdatei !';
        CtrlVisible[edt16] := True;
        if (FDialogMode = dmOpen) then CtrlEnabled[IDOK] := False;
        Exit;
      end;
      CtrlVisible[edt16] := False;
      CtrlEnabled[ID_OK] := True;
      CtrlText[grp4] := 'Textdatei (Pattern-, Recallset)';
      CtrlText[stc10] := 'Muster';
      CtrlText[stc12] := 'Ausgnge';
      try
        CtrlText[edt10] := ReadInt;
        CtrlText[edt11] := ReadInt;
        CtrlText[edt12] := ReadInt;
      except
        CtrlText[edt16] := 'Kein gltiger Integerwert !';
        CtrlVisible[edt16] := True;
        if (FDialogMode = dmOpen) then CtrlEnabled[IDOK] := False;
      end;
      if Assigned(FNeuralNet) then begin
        CtrlText[edt13] := IntToStr(FNeuralNet.PatternSet.PatternNo);
        CtrlText[edt14] := IntToStr(FNeuralNet.PatternSet.InputNo);
        CtrlText[edt15] := IntToStr(FNeuralNet.PatternSet.OutputNo);
      end;
    finally
      CloseFile(AscFile);
    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('Dialoge', [TFileDialog]);
  RegisterPropertyEditor(TypeInfo(string), TFileDialog, 'Filter', TFilterProperty);
  RegisterComponents('NeuroNet', [TNetFileDialog]);
end;

end.
